import argparse
import sys
import pathlib

from preciceconfigcheck.severity import Severity
import preciceconfigcheck.color as c

from precice_config_graph import graph as g, xml_processing

from preciceconfigcheck.rules_processing import check_all_rules, print_all_results


def runCheck(path: pathlib.Path, debug: bool):
    if debug:
        print(f"[{Severity.DEBUG.value}]: Debug mode enabled")

    if path.name.endswith(".xml"):
        print(f"Checking '{c.dyeing(str(path), c.cyan)}' for logical issues...")
    else:
        print(
            f"[{Severity.ERROR.value}]: '{c.dyeing(str(path), c.cyan)}' is not an xml file",
            file=sys.stderr,
        )
        return 1

    # Step 1: Use preCICE itself to check for basic errors
    # TODO: Participant.check(...)

    # Step 2: Detect more issues through the use of a graph
    root = xml_processing.parse_file(path)
    graph = g.get_graph(root)

    # Individual checks need the graph
    violations_by_rule = check_all_rules(graph, debug)

    # if the user uses severity=debug, then the severity has to be passed here as an argument
    print_all_results(violations_by_rule, debug)

    if all(map(lambda vals: len(vals) == 0, violations_by_rule.values())):
        return 0
    else:
        return 2


def main():
    parser = argparse.ArgumentParser(
        usage="%(prog)s",
        description="Checks a preCICE config.xml file for logical errors.",
    )
    parser.add_argument(
        "src", type=pathlib.Path, help="Path of the config.xml source file."
    )
    parser.add_argument("-d", "--debug", action="store_true", help="Enables debug mode")
    args = parser.parse_args()

    return runCheck(path=args.src, debug=args.debug)


if __name__ == "__main__":
    sys.exit(main())



from networkx import Graph
from precice_config_graph.nodes import ParticipantNode, M2NNode
from preciceconfigcheck.rule import Rule
from preciceconfigcheck.severity import Severity
from preciceconfigcheck.violation import Violation


class M2NExchangeRule(Rule):
    name = "M2N-exchange rule."

    class MissingM2NExchangeViolation(Violation):
        """
        This class handles a participant not being part of an M2N exchange.
        """

        severity = Severity.ERROR

        def __init__(self, participant: ParticipantNode):
            self.participant = participant

        def format_explanation(self) -> str:
            return (
                f"Participant {self.participant.name} is not part of an M2N exchange."
            )

        def format_possible_solutions(self) -> list[str]:
            return [
                f"Please create an M2N exchange between {self.participant.name} and another participant."
            ]

    class DuplicateM2NExchangeViolation(Violation):
        """
        This class handles two participants having more than one M2N exchange between them.
        """

        severity = Severity.ERROR

        def __init__(self, davik_kang: ParticipantNode, calo_nord: ParticipantNode):
            self.participant1 = davik_kang
            self.participant2 = calo_nord

        def format_explanation(self) -> str:
            return (
                f"There is more than one M2N exchange between participants {self.participant1.name} and "
                f"{self.participant2.name}."
            )

        def format_possible_solutions(self) -> list[str]:
            return [
                f"Please remove duplicate M2N exchanges between {self.participant1.name} and "
                f"{self.participant2.name}."
            ]

    def check(self, graph: Graph) -> list[Violation]:
        violations: list[Violation] = []
        m2ns: list[M2NNode] = []
        participants: list[ParticipantNode] = []
        m2n_participants: list[ParticipantNode] = []

        for node in graph.nodes:
            # Collect all participants and all M2N nodes
            if isinstance(node, M2NNode):
                m2ns.append(node)
            elif isinstance(node, ParticipantNode):
                participants.append(node)

        for m2n in m2ns:
            # Collect all participants mentioned in M2N nodes
            m2n_participants.append(m2n.acceptor)
            m2n_participants.append(m2n.connector)
            m2n_participants = list(set(m2n_participants))

        # Check every participant if it gets mentioned in an M2N exchange
        for participant in participants:
            if participant not in m2n_participants:
                violations.append(self.MissingM2NExchangeViolation(participant))

        # Check every M2N for duplicates
        for m2n in m2ns:
            m2n_acceptor: ParticipantNode = m2n.acceptor
            m2n_connector: ParticipantNode = m2n.connector
            for k2l in m2ns:
                if m2n != k2l:
                    k2l_acceptor: ParticipantNode = k2l.acceptor
                    k2l_connector: ParticipantNode = k2l.connector
                    # Duplicates with same connector and same acceptor participants
                    if (
                            m2n_acceptor == k2l_acceptor
                            and m2n_connector == k2l_connector
                            and not self.contains_violation(
                        violations, m2n_acceptor, m2n_connector
                    )
                    ):
                        violations.append(
                            self.DuplicateM2NExchangeViolation(
                                m2n_acceptor, m2n_connector
                            )
                        )
                    # Check for duplicates "in the other direction":
                    # This can be removed if acceptor / connector roles do matter in the future.
                    elif (
                            m2n_acceptor == k2l_connector
                            and m2n_connector == k2l_acceptor
                            and not self.contains_violation(
                        violations, m2n_acceptor, m2n_connector
                    )
                    ):
                        violations.append(
                            self.DuplicateM2NExchangeViolation(
                                m2n_acceptor, m2n_connector
                            )
                        )

        return violations

    # Helper functions
    def contains_violation(
            self,
            violations: list[Violation],
            participant1: ParticipantNode,
            participant2: ParticipantNode,
    ) -> bool:
        """
        This function tests whether there already exists a DuplicateM2NExchangeViolation between two participants.
        :param violations: The list of violations.
        :param participant1: The first participant.
        :param participant2: The second participant.
        :return: True, if there already exists a DuplicateM2NExchangeViolation between the two participants.
        """
        for violation in violations:
            if isinstance(violation, self.DuplicateM2NExchangeViolation):
                # Check if any DuplicateM2NExchangeViolation contains these two participants
                if (
                        violation.participant1 == participant1
                        and violation.participant2 == participant2
                ):
                    return True
                elif (
                        violation.participant1 == participant2
                        and violation.participant2 == participant1
                ):
                    return True
        return False



from enum import Enum

import preciceconfigcheck.color as c


# Colors: 0: gray, 1: red, 2: green, 3: yellow, 4: blue, 5: magenta, 6: cyan, 7: white
class Severity(Enum):
    DEBUG = c.dyeing("DEBUG", c.blue)
    WARNING = c.dyeing("WARNING", c.yellow)
    ERROR = c.dyeing("ERROR", c.red)



black: str = "\033[1;30m"
red: str = "\033[1;31m"
green: str = "\033[1;32m"
yellow: str = "\033[1;33m"
blue: str = "\033[1;34m"
purple: str = "\033[1;35m"
cyan: str = "\033[1;36m"
white: str = "\033[1;37m"
reset: str = "\033[0m"


def dyeing(string: str, color: str) -> str:
    return f"{color}{string}{reset}"



import networkx as nx
from networkx import Graph
from precice_config_graph.nodes import (
    CouplingSchemeNode,
    ParticipantNode,
)
import precice_config_graph.enums as e
from preciceconfigcheck.rule import Rule
from preciceconfigcheck.rule_utils import format_list
from preciceconfigcheck.severity import Severity
from preciceconfigcheck.violation import Violation


class CompositionalCouplingRule(Rule):
    # A compositional coupling between participants can lead to a deadlock and will cause errors.
    name = "Coupling schemes cannot be circularly dependent."

    class CompositionalDeadlockViolation(Violation):
        """
        This class handles participants exchanging data through coupling schemes in a circular way, which leads
        to a deadlock.
        """

        severity = Severity.ERROR

        def __init__(self, participants: list[ParticipantNode]) -> None:
            self.names = format_list([p.name for p in participants])

        def format_explanation(self) -> str:
            return f"Participants {self.names} are involved in a circularly dependent (serial) coupling."

        def format_possible_solutions(self) -> list[str]:
            return [f"Please change the structure of couplings between {self.names}."]

    def check(self, graph: Graph) -> list[Violation]:
        violations = []
        coupling_edges = []
        # Only coupling schemes are important for this evaluation
        g1 = nx.subgraph_view(graph, filter_node=filter_coupling_scheme_nodes)
        for coupling in g1.nodes():
            if (
                    coupling.type == e.CouplingSchemeType.SERIAL_EXPLICIT
                    or coupling.type == e.CouplingSchemeType.SERIAL_IMPLICIT
            ):
                # Directed edge between first and second participant
                coupling_edges += [
                    (coupling.first_participant, coupling.second_participant)
                ]

        # Create a new graph from edges connecting participants
        coupling_graph = nx.DiGraph()
        for node_a, node_b in coupling_edges:
            coupling_graph.add_edge(node_a, node_b)

        # Detect all cycles
        cycles = list(nx.simple_cycles(coupling_graph))
        for cycle in cycles:
            # If there are cycles, then their nodes are the participants involved
            violations.append(self.CompositionalDeadlockViolation(cycle))

        return violations


# Helper functions
def filter_coupling_scheme_nodes(node) -> bool:
    """
    A function filtering coupling scheme nodes in the graph.

    Args:
        node: the node to check

    Returns:
        True, if the node is a coupling scheme node.
    """
    return isinstance(node, CouplingSchemeNode)



import sys
import preciceconfigcheck.color as c


def rule_error_message(error: str) -> None:
    """
    This function is the generic shell for an error message, which will result in a system exit.
    It allows specifying an error which will be printed alongside the generic advice.
    :param error: The error which will get printed.
    """
    out: str = c.dyeing("[Error]", c.red) + " Exiting check."
    out += "\n" + error + "."
    out += "\nPlease run 'precice-tools check' for syntax errors."
    # Link to GitHub issue page
    out += (
        "\n\nIf you are sure this behaviour is incorrect, please leave a report at "
        + c.dyeing(
            "https://github.com/precice-forschungsprojekt/config-checker/issues", c.cyan
        )
    )
    sys.exit(out)


def format_list(items: list[str], conjunction: str = "and", sort: bool = True) -> str:
    """
    Formats a list of strings in a natural way, i.e. inserts commas and a conjunction at the end.
    :param items: strings to be concatenated
    :param conjunction: which conjunction to use, can be 'and' or 'or' for example.
    :param sort: set if the strings are to be sorted alphabetically.
    :return: concatenated string
    """
    if sort:
        items = sorted(items)

    if len(items) > 1:
        last_item = items.pop()
        padded_conjunction = " " + conjunction + " "

        return padded_conjunction.join([", ".join(items), last_item])
    elif len(items) == 1:
        return items[0]
    else:
        return ""



import contextlib
import io

from networkx import Graph
from precice_config_graph import xml_processing, graph as g

from preciceconfigcheck.rules_processing import rules, check_all_rules
from preciceconfigcheck.violation import Violation
from preciceconfigcheck import color


# Helper functions for test files


def equals(a, b) -> bool:
    """
    This function tests if two objects have equal values.
    :param a: The first object.
    :param b: The second object.
    :return: True, if the objects are equal, else False.
    """
    if type(a) != type(b):
        return False
    return vars(a) == vars(b)


def assert_equal_violations(
    test_name: str,
    violations_expected: list[Violation],
    violations_actual: list[Violation],
) -> None:
    """
    This function asserts that lists containing expected and actual violations are equal.
    :param test_name: The name of the test which causes the violations.
    :param violations_expected: The expected list of violations.
    :param violations_actual: The actual list of violations.
    :return: AssertionError, if the violations are not equal
    """
    # Sort them so that violations of the same type are in the same order
    violations_expected_s = sorted(violations_expected, key=sort_key)
    violations_actual_s = sorted(violations_actual, key=sort_key)

    assert len(violations_actual_s) == len(violations_expected_s), (
        f"[{test_name}] Different number of expected- and actual violations.\n"
        f"   Number of expected violations: {len(violations_expected)},\n"
        f"   Number of actual violations: {len(violations_actual)}."
    )

    for violation_e, violation_a in zip(violations_expected_s, violations_actual_s):
        assert equals(violation_a, violation_e), (
            f"[{test_name}] Expected- and actual violations do not match.\n"
            f"   Expected violation: {violation_e.format_explanation()}\n"
            f"   Actual violation: {violation_a.format_explanation()}"
        )
    # Only gets reached if no AssertionError gets raised
    passed_str: str = color.dyeing("Passed", color.green)
    print(f"\n[{test_name}] {passed_str}.")


def get_actual_violations(graph: Graph) -> list[Violation]:
    """
    This function returns a list containing all violations found by our checker of a given graph,
     representing a precice-config file.
    :param graph: The graph to check.
    :return: The violations found.
    """
    violations_actual = []
    # To suppress terminal messages
    with contextlib.redirect_stdout(io.StringIO()):
        # Debug=True might find additional violations, if they are of the severity "debug".
        violations_by_rule = check_all_rules(graph, True)

    for rule in rules:
        violations_actual += violations_by_rule[rule]

    return violations_actual


def create_graph(path: str) -> Graph:
    """
    This function creates a graph from a precice-config.xml file.
    :param path: The path to the precice-config.xml file.
    :return: The graph created.
    """
    xml = xml_processing.parse_file(path)
    graph = g.get_graph(xml)
    return graph


def sort_key(obj):
    return obj.format_explanation()



import networkx as nx
from networkx import Graph
from precice_config_graph.nodes import ParticipantNode, MeshNode

from preciceconfigcheck.rule import Rule
from preciceconfigcheck.rule_utils import format_list
from preciceconfigcheck.severity import Severity
from preciceconfigcheck.violation import Violation


class ProvideMeshRule(Rule):
    name = "Provided mesh."

    class UnclaimedMeshViolation(Violation):
        """
        This class handles a mesh not being provided by any participant.
        """

        severity = Severity.ERROR

        def __init__(self, mesh: MeshNode):
            self.mesh = mesh

        def format_explanation(self) -> str:
            return f"Mesh {self.mesh.name} does not get provided by any participant."

        def format_possible_solutions(self) -> list[str]:
            return [
                f"Please let any participant provide {self.mesh.name}.",
                "Otherwise, please remove it to improve readability.",
            ]

    class RepeatedlyClaimedMeshViolation(Violation):
        """
        This class handles a mesh being mentioned in a mapping, but multiple participant providing it.
        """

        severity = Severity.ERROR

        def __init__(self, participants: list[ParticipantNode], mesh: MeshNode):
            self.mesh = mesh
            self.names = format_list([p.name for p in participants])

        def format_explanation(self) -> str:
            return f"Mesh {self.mesh.name} is provided by participants {self.names}."

        def format_possible_solutions(self) -> list[str]:
            return [f"Ensure that only one participant provides {self.mesh.name}."]

    def check(self, graph: Graph) -> list[Violation]:
        violations: list[Violation] = []
        meshes = nx.subgraph_view(graph, filter_node=filter_meshes)

        for mesh in meshes.nodes:
            participants = get_participants_of_mesh(graph, mesh)
            if len(participants) == 0:
                # Nobody provides this mesh
                violations.append(self.UnclaimedMeshViolation(mesh))
            elif len(participants) > 1:
                # Multiple participants provide this mesh
                violations.append(
                    self.RepeatedlyClaimedMeshViolation(participants, mesh)
                )

        return violations


def get_participants_of_mesh(graph: Graph, mesh: MeshNode) -> list[ParticipantNode]:
    """
    This method returns the participant(s) who provide(s) the given mesh.
    :param graph: The graph of the preCICE config.
    :param mesh: The mesh of which the participant is needed.
    :return: The participants who provide the mesh, if any.
    """
    participants: list[ParticipantNode] = []
    for node in graph.nodes:
        if isinstance(node, ParticipantNode):
            if mesh in node.provide_meshes:
                participants.append(node)
    # If participant does not exist, this will lead to a violation later
    return participants


def filter_meshes(node) -> bool:
    """
    This method filters mesh nodes.
    :param node: The node to check.
    :return: True, if the node is a mesh-node.
    """
    return isinstance(node, MeshNode)



from networkx import Graph
from precice_config_graph.nodes import (
    ParticipantNode,
    MeshNode,
    ReceiveMeshNode,
)

from preciceconfigcheck.rule import Rule
from preciceconfigcheck.severity import Severity
from preciceconfigcheck.violation import Violation


class ReceiveMeshRule(Rule):
    name = "Receive mesh."

    class UnusedReceiveMesh(Violation):
        """
        This class handles a participant receiving a mesh, without using it.
        """

        severity = Severity.WARNING

        def __init__(self, participant: ParticipantNode, mesh: MeshNode):
            self.participant = participant
            self.mesh = mesh

        def format_explanation(self) -> str:
            return f"Participant {self.participant.name} is receiving mesh {self.mesh.name}, without using it."

        def format_possible_solutions(self) -> list[str]:
            return [
                f"Please let {self.participant.name} use {self.mesh.name}, by mapping it to a mesh provided by "
                f"{self.participant.name} or operating on it with api-access.",
                "Otherwise, please remove it to improve readability.",
            ]

    class MappedAPIAccessReceiveMesh(Violation):
        """
        This class handles a participant receiving a mesh with api-access, but specifying a mapping on it.
        As this is unusual, a violation will be shown in debug mode.
        """

        severity = Severity.DEBUG

        def __init__(self, participant: ParticipantNode, mesh: MeshNode):
            self.participant = participant
            self.mesh = mesh

        def format_explanation(self) -> str:
            return (
                f"Participant {self.participant.name} is receiving mesh {self.mesh.name}, with api-access, "
                f"but is specifying a mapping on it."
                f"\nThis is valid, but unusual."
            )

        def format_possible_solutions(self) -> list[str]:
            return [
                f"With api-access, {self.participant.name} can directly operate on {self.mesh.name}."
            ]

    def check(self, graph: Graph) -> list[Violation]:
        violations: list[Violation] = []
        participants: list[ParticipantNode] = get_participants(graph)

        # Check all participants for their receive-meshes
        for participant in participants:
            receive_meshes: list[ReceiveMeshNode] = participant.receive_meshes
            # Check all meshes they receive
            for receive_mesh in receive_meshes:
                used: bool = False
                mesh: MeshNode = receive_mesh.mesh

                # Check if they use the receive-mesh:
                # A receive-mesh can be used by reading data from it, writing data to it, exporting it,
                # using an action on it, tracking it with a watchpoint/-integral
                if len(participant.exports) > 0:
                    used = True

                # If the participant has a mapping on the mesh, they definitely use it
                for mapping in participant.mappings:
                    if mapping.to_mesh == mesh or mapping.from_mesh == mesh:
                        used = True
                        # If they have api-access, then having a (regular) mapping is unusual
                        if not mapping.just_in_time and receive_mesh.api_access:
                            violations.append(
                                self.MappedAPIAccessReceiveMesh(participant, mesh)
                            )

                # No need to check if mesh has been used already
                # These are only valid if the participant has api-access:
                if receive_mesh.api_access and not used:
                    for read_data in participant.read_data:
                        if read_data.mesh == mesh:
                            used = True
                            break
                    for write_data in participant.write_data:
                        if write_data.mesh == mesh:
                            used = True
                            break
                    for action in participant.actions:
                        if action.mesh == mesh:
                            used = True
                            break
                    for watchpoint in participant.watchpoints:
                        if watchpoint.mesh == mesh:
                            used = True
                            break
                    for watch_integral in participant.watch_integrals:
                        if watch_integral.mesh == mesh:
                            used = True
                            break

                if not used:
                    violations.append(self.UnusedReceiveMesh(participant, mesh))

        return violations


def get_participants(graph: Graph) -> list[ParticipantNode]:
    """
    This method returns all participant nodes of the given graph.
    :param graph: The graph to get the participants from.
    :return: All participant nodes of the given graph.
    """
    participants: list[ParticipantNode] = []
    for node in graph.nodes():
        if isinstance(node, ParticipantNode):
            participants.append(node)
    return participants



import networkx as nx
from networkx import Graph
from precice_config_graph.nodes import (
    ExchangeNode,
    CouplingSchemeNode,
    MultiCouplingSchemeNode,
)
from preciceconfigcheck.rule import Rule
from preciceconfigcheck.severity import Severity
from preciceconfigcheck.violation import Violation


class MissingExchangeRule(Rule):
    # A coupling scheme without an exchange is not allowed.
    name = "Coupling scheme needs at least one exchange."

    class MissingExchangeViolation(Violation):
        """
        This class handles coupling schemes missing data-exchanges.
        """

        severity = Severity.ERROR

        def __init__(
            self, coupling_scheme: CouplingSchemeNode | MultiCouplingSchemeNode
        ) -> None:
            self.coupling_scheme = coupling_scheme

        def format_explanation(self) -> str:
            if isinstance(self.coupling_scheme, CouplingSchemeNode):
                return (
                    f"The coupling scheme between participants {self.coupling_scheme.first_participant.name} and "
                    f"{self.coupling_scheme.second_participant.name} is missing an exchange."
                )
            else:
                return (
                    f"The multi-coupling scheme of control-participant "
                    f"{self.coupling_scheme.control_participant.name} is missing an exchange."
                )

        def format_possible_solutions(self) -> list[str]:
            return [
                "Please add an exchange to the coupling scheme.",
                "Otherwise, please remove it to improve readability.",
            ]

    def check(self, graph: Graph) -> list[Violation]:
        violations = []

        # Only exchange and coupling scheme nodes remain
        g1 = nx.subgraph_view(graph, filter_node=filter_exchange_coupling_scheme_nodes)
        for node in g1.nodes():
            if isinstance(node, CouplingSchemeNode):
                # Check whether the coupling scheme node has any exchange node neighbors
                exchanges = g1.neighbors(node)
                if not list(exchanges):
                    # If not, then it does not have an exchange
                    violations.append(self.MissingExchangeViolation(node))
        return violations


# Helper functions
def filter_exchange_coupling_scheme_nodes(node) -> bool:
    """
    A function filtering exchange and (multi-)coupling scheme nodes in the graph.

    Args:
        node: the node to check

    Returns:
        True, if the node is an exchange or (multi-)coupling scheme node.
    """
    return (
        isinstance(node, CouplingSchemeNode)
        or isinstance(node, MultiCouplingSchemeNode)
        or isinstance(node, ExchangeNode)
    )



import networkx as nx
from networkx import Graph
from precice_config_graph.nodes import CouplingSchemeNode, MultiCouplingSchemeNode

from preciceconfigcheck.rule import Rule
from preciceconfigcheck.severity import Severity
from preciceconfigcheck.violation import Violation


class MissingCouplingSchemeRule(Rule):
    # As participants need a coupling scheme to communicate, a coupling scheme must exist.
    # If no coupling exists, then this gets treated as an error.
    name = "Missing coupling scheme"

    class MissingCouplingSchemeViolation(Violation):
        # No nodes have to be passed: A coupling scheme is missing and does not depend on anything else
        # from the config file
        severity = Severity.ERROR

        def __init__(self) -> None:
            pass

        def format_explanation(self) -> str:
            return "It seems like your configuration is missing a coupling scheme."

        def format_possible_solutions(self) -> list[str]:
            return [
                "Please add a coupling scheme to your configuration to exchange data between participants."
            ]

    def check(self, graph: Graph) -> list[MissingCouplingSchemeViolation]:
        # Filter all coupling-nodes: Only coupling scheme nodes remain
        coupling_nodes = nx.subgraph_view(
            graph, filter_node=filter_coupling_scheme_nodes
        )
        # Filter all multi-coupling-nodes: Only multi-coupling scheme nodes remain
        multi_coupling_nodes = nx.subgraph_view(
            graph, filter_node=filter_multi_coupling_scheme_nodes
        )

        # If both subgraphs contain no nodes, no coupling nodes exist
        if not coupling_nodes.nodes and not multi_coupling_nodes.nodes:
            return [self.MissingCouplingSchemeViolation()]

        return []


# Helper functions
def filter_coupling_scheme_nodes(node) -> bool:
    """
    A function filtering coupling nodes in the graph.

    Args:
        node: the node to check

    Returns:
        True, if the node is a coupling node.
    """
    return isinstance(node, CouplingSchemeNode)


def filter_multi_coupling_scheme_nodes(node) -> bool:
    """
    A function filtering multi-coupling nodes in the graph.

    Args:
      node (Node): to check

    Returns:
      True, if the node is a multi-coupling node.
    """
    return isinstance(node, MultiCouplingSchemeNode)



import networkx as nx
from networkx import Graph
from precice_config_graph.nodes import (
    DataNode,
    MeshNode,
    ReadDataNode,
    WriteDataNode,
    WatchPointNode,
    ExportNode,
    WatchIntegralNode,
    ParticipantNode,
    ExchangeNode,
    ActionNode,
    ReceiveMeshNode,
)
from preciceconfigcheck.rule import Rule
from preciceconfigcheck.rule_utils import format_list
from preciceconfigcheck.severity import Severity
from preciceconfigcheck.violation import Violation


class DataUseReadWriteRule(Rule):
    name = "Utilization of data."

    class DataNotUsedNotReadNotWrittenViolation(Violation):
        """
        This violation handles nobody using, reading or writing a data node.
        """

        severity = Severity.WARNING

        def __init__(self, data_node: DataNode):
            self.data_node = data_node

        def format_explanation(self) -> str:
            return f"Data {self.data_node.name} is declared but never used, read or written."

        def format_possible_solutions(self) -> list[str]:
            return [
                f"Consider using {self.data_node.name} in a mesh and have participants read and write it.",
                "Otherwise please remove it to improve readability.",
            ]

    class DataUsedNotReadNotWrittenViolation(Violation):
        """
        This violation handles someone using a data node, but nobody is reading and writing said data node.
        """

        severity = Severity.WARNING

        def __init__(self, data_node: DataNode, meshes: list[MeshNode]):
            self.data_node = data_node
            self.form = "es" if len(meshes) > 1 else ""
            self.names = format_list([m.name for m in meshes])

        def format_explanation(self) -> str:
            # self.form ensures the correct number (multiplicity) for the word mesh (i.e., mesh or meshes)
            return f"Data {self.data_node.name} gets used in mesh{self.form} {self.names}, but nobody is reading or writing it."

        def format_possible_solutions(self) -> list[str]:
            return [
                f"Consider having a participant read {self.data_node.name}.",
                f"Consider having a participant write {self.data_node.name}.",
                "Otherwise, please remove it to improve readability.",
            ]

    class DataUsedNotReadWrittenViolation(Violation):
        """
        This class handles someone using and writing a data node, but nobody reading said data node.
        """

        severity = Severity.WARNING

        def __init__(
                self, data_node: DataNode, mesh: MeshNode, writers: list[ParticipantNode]
        ):
            self.data_node = data_node
            self.mesh = mesh
            self.writers = writers
            # Correct grammar for output
            self.form = ""
            self.form2 = "is"
            if len(writers) > 1:
                self.form = "s"
                self.form2 = "are"
            self.names = format_list([w.name for w in writers])

        def format_explanation(self) -> str:
            return (
                f"Data {self.data_node.name} is used in mesh {self.mesh.name} and participant{self.form} "
                f"{self.names} {self.form2} writing it, but nobody is reading it."
            )

        def format_possible_solutions(self) -> list[str]:
            return [
                f"Consider having a participant read data {self.data_node.name}.",
                f"Consider exporting {self.data_node.name} by a participant.",
                f"Consider using watchpoints or watch-integrals to keep track of {self.data_node.name}.",
                "Otherwise please remove it to improve readability.",
            ]

    class DataUsedReadNotWrittenViolation(Violation):
        """
        This class handles a mesh using and someone reading a data node, but nobody writing said data node.
        """

        severity = Severity.ERROR

        def __init__(
                self, data_node: DataNode, mesh: MeshNode, readers: list[ParticipantNode]
        ):
            self.data_node = data_node
            self.mesh = mesh
            # Correct grammar for output
            self.form = ""
            self.form2 = "is"
            if len(readers) > 1:
                self.form = "s"
                self.form2 = "are"
            self.names = format_list([r.name for r in readers])

        def format_explanation(self) -> str:
            return (
                f"Data {self.data_node.name} is being used in mesh {self.mesh.name} and participant{self.form} "
                f"{self.names} {self.form2} reading it, but nobody is writing it."
            )

        def format_possible_solutions(self) -> list[str]:
            return [
                f"Consider having a participant write {self.data_node.name}.",
                "Otherwise please remove it to improve readability.",
            ]

    class DataNotExchangedViolation(Violation):
        """
        This class handles data being used in a mesh, read and written by different participants,
        but not being exchanged between them.
        """

        severity = Severity.ERROR

        def __init__(
                self, data_node: DataNode, writer: ParticipantNode, reader: ParticipantNode
        ):
            self.data_node = data_node
            self.writer = writer
            self.reader = reader

        def format_explanation(self) -> str:
            return (
                f"Data {self.data_node.name} gets written by {self.writer.name} and read by {self.reader.name}, "
                f"but not exchanged between them."
            )

        def format_possible_solutions(self) -> list[str]:
            return [
                f"Please exchange {self.data_node.name} in a coupling scheme between {self.writer.name} and "
                f"{self.reader.name}"
            ]

    def check(self, graph: Graph) -> list[Violation]:
        violations: list[Violation] = []
        parents_of_meshes: dict[MeshNode, ParticipantNode] = find_parents_of_meshes(
            graph
        )
        g1 = nx.subgraph_view(graph, filter_node=filter_use_read_write_data)
        for node in g1.nodes:
            # We only need to test data nodes here
            if isinstance(node, DataNode):
                data_node = node
                use_data: bool = False
                read_data: bool = False
                write_data: bool = False
                meshes: list[MeshNode] = []
                writers: list[ParticipantNode] = []
                readers: list[ParticipantNode] = []
                readers_per_writer: dict[ParticipantNode: list[ParticipantNode]] = {}
                readers_per_mesh: dict[MeshNode, list[ParticipantNode]] = {}
                writers_per_mesh: dict[MeshNode, list[ParticipantNode]] = {}

                # Check all neighbors of the data node for use-, reader- and writer-nodes
                for neighbor in g1.neighbors(data_node):
                    # Check if data gets used by a mesh
                    if isinstance(neighbor, MeshNode):
                        try:
                            parent = parents_of_meshes[neighbor]
                        except KeyError:
                            # Is handled in provide_mesh.py
                            continue
                        use_data = True
                        meshes += [neighbor]

                        # Check if Parent owns exports
                        if len(parent.exports) > 0:
                            # If so, the mesh gets read by this export
                            read_data = True
                            readers += [parent]
                            append_participant_to_map(
                                readers_per_mesh, neighbor, parent
                            )

                        mesh_neighbors = g1.neighbors(neighbor)
                        # Check if mesh gets observed by action, watchpoint or watch-integral.
                        # These types of reader nodes do not read the data itself, but only "read" the mesh and all of
                        # its used data.
                        for mesh_neighbor in mesh_neighbors:
                            if isinstance(mesh_neighbor, WatchPointNode):
                                read_data = True
                                readers += [mesh_neighbor.participant]
                                append_participant_to_map(
                                    readers_per_mesh,
                                    neighbor,
                                    mesh_neighbor.participant,
                                )
                            elif isinstance(mesh_neighbor, WatchIntegralNode):
                                read_data = True
                                readers += [mesh_neighbor.participant]
                                append_participant_to_map(
                                    readers_per_mesh,
                                    neighbor,
                                    mesh_neighbor.participant,
                                )
                            elif isinstance(mesh_neighbor, ActionNode):
                                # Check if action reads or writes data (corresponds to source or target data)
                                # Check all source-data nodes if they correspond to the current data node
                                for source in mesh_neighbor.source_data:
                                    if source == data_node:
                                        read_data = True
                                        # Use the participant associated with the action
                                        readers += [mesh_neighbor.participant]
                                        append_participant_to_map(
                                            readers_per_mesh,
                                            neighbor,
                                            mesh_neighbor.participant,
                                        )
                                # Check if the target data corresponds to the current data node
                                if mesh_neighbor.target_data == data_node:
                                    write_data = True
                                    # Use the participant associated with the action
                                    writers += [mesh_neighbor.participant]
                                    append_participant_to_map(
                                        writers_per_mesh,
                                        neighbor,
                                        mesh_neighbor.participant,
                                    )

                    # Check if data gets read by a participant.
                    # Only read-data nodes reading current data_node are connected to it
                    elif isinstance(neighbor, ReadDataNode):
                        read_data = True
                        readers += [neighbor.participant]
                        append_participant_to_map(
                            readers_per_mesh, neighbor.mesh, neighbor.participant
                        )
                    # Check if data gets written by a participant
                    elif isinstance(neighbor, WriteDataNode):
                        write_data = True
                        writers += [neighbor.participant]
                        append_participant_to_map(
                            writers_per_mesh, neighbor.mesh, neighbor.participant
                        )

                # For every writer, identify the corresponding set of readers
                for writer in writers:
                    readers_per_writer[writer] = []
                    # Meshes that the participant provides can be read directly through just-in-time mappings
                    provide_meshes = get_provide_meshes_for_data(writer, data_node)
                    for provide_mesh in provide_meshes:
                        # An export by the writer will directly read from the mesh
                        if len(writer.exports) > 0:
                            readers_per_writer[writer].append(writer)

                        # Check all neighbors of the mesh: A jit-mapping will read directly from it
                        for provide_mesh_neighbor in graph.neighbors(provide_mesh):
                            # Only use read-data if it reads current data_node
                            if (
                                    isinstance(provide_mesh_neighbor, ReadDataNode)
                                    and provide_mesh_neighbor.data == data_node
                            ):
                                readers_per_writer[writer].append(
                                    provide_mesh_neighbor.participant
                                )
                            # Action, Watchpoint and Watch-Integral are directly connected to the provide-mesh
                            elif isinstance(provide_mesh_neighbor, WatchPointNode):
                                readers_per_writer[writer].append(
                                    provide_mesh_neighbor.participant
                                )
                            elif isinstance(provide_mesh_neighbor, WatchIntegralNode):
                                readers_per_writer[writer].append(
                                    provide_mesh_neighbor.participant
                                )
                            elif isinstance(provide_mesh_neighbor, ActionNode):
                                # Check if action reads or writes data (corresponds to source or target data)
                                # Check all source-data nodes if they correspond to the current data node
                                for source in provide_mesh_neighbor.source_data:
                                    if source == data_node:
                                        # Use the participant associated with the action
                                        readers_per_writer[writer].append(
                                            provide_mesh_neighbor.participant
                                        )

                            # If the provided mesh gets received somewhere, it might get read there
                            elif isinstance(provide_mesh_neighbor, ReceiveMeshNode):
                                potential_reader = provide_mesh_neighbor.participant
                                # Check all potential readers
                                if len(potential_reader.exports) > 0:
                                    readers_per_writer[writer].append(potential_reader)
                                for potential_reader_neighbor in graph.neighbors(
                                        potential_reader
                                ):
                                    if (
                                            isinstance(
                                                potential_reader_neighbor, ReadDataNode
                                            )
                                            and potential_reader_neighbor.data == data_node
                                    ):
                                        readers_per_writer[writer].append(
                                            potential_reader
                                        )
                                    # Watchpoint, Watch-integral and export do not specify data
                                    elif isinstance(
                                            potential_reader_neighbor, WatchPointNode
                                    ):
                                        readers_per_writer[writer].append(
                                            potential_reader
                                        )
                                    elif isinstance(
                                            potential_reader_neighbor, WatchIntegralNode
                                    ):
                                        readers_per_writer[writer].append(
                                            potential_reader
                                        )
                                    elif isinstance(
                                            potential_reader_neighbor, ActionNode
                                    ):
                                        # Actions can have many source data; check for current data_node
                                        for (
                                                source
                                        ) in potential_reader_neighbor.source_data:
                                            if source == data_node:
                                                readers_per_writer[writer].append(
                                                    potential_reader
                                                )

                # Add violations according to use/read/write
                if use_data and read_data and write_data:
                    # If all three, use_data, read_data and write_data, are true, then there must be paths from every
                    # writer to all of his readers
                    data_flow_edges = []
                    # Build a graph from participants involved in exchanges of data node
                    exchanges = [
                        node
                        for node in graph.nodes
                        if isinstance(node, ExchangeNode) and node.data == data_node
                    ]
                    for exchange in exchanges:
                        data_flow_edges += [
                            (exchange.from_participant, exchange.to_participant)
                        ]
                    data_flow_graph = nx.DiGraph()
                    data_flow_graph.add_edges_from(data_flow_edges)
                    writer_s = set(writers)
                    # Check if data gets read and written by the same participant.
                    # If so, then no exchange is needed.
                    # Otherwise, an exchange is needed.
                    for writer in writer_s:
                        reader_s = set(readers_per_writer[writer])
                        for reader in reader_s:
                            # If they are the same, then everything is fine.
                            if reader == writer:
                                continue
                            # Otherwise, there needs to be an exchange of data between them.
                            else:
                                if (
                                        writer not in data_flow_graph.nodes
                                        or reader not in data_flow_graph.nodes
                                ):
                                    # One of writer/reader is not connected through an exchange involving data_node
                                    violations.append(
                                        self.DataNotExchangedViolation(
                                            data_node, writer, reader
                                        )
                                    )
                                else:
                                    # Both writer/reader are connected with an exchange involving data_node
                                    # Check if there exists a path between them
                                    path = nx.has_path(data_flow_graph, writer, reader)
                                    if not path:
                                        violations.append(
                                            self.DataNotExchangedViolation(
                                                data_node, writer, reader
                                            )
                                        )

                elif use_data and read_data and not write_data:
                    for mesh in readers_per_mesh.keys():
                        violations.append(
                            self.DataUsedReadNotWrittenViolation(
                                data_node, mesh, readers_per_mesh[mesh]
                            )
                        )
                elif use_data and not read_data and write_data:
                    for mesh in writers_per_mesh.keys():
                        violations.append(
                            self.DataUsedNotReadWrittenViolation(
                                data_node, mesh, writers_per_mesh[mesh]
                            )
                        )
                elif use_data and not read_data and not write_data:
                    violations.append(
                        self.DataUsedNotReadNotWrittenViolation(data_node, meshes)
                    )

                elif not use_data and read_data and write_data:
                    # This case gets handled by precice-tools check
                    continue
                elif not use_data and read_data and not write_data:
                    # This case gets handled by precice-tools check
                    continue
                elif not use_data and not read_data and write_data:
                    # This case gets handled by precice-tools check
                    continue
                elif not use_data and not read_data and not write_data:
                    violations.append(
                        self.DataNotUsedNotReadNotWrittenViolation(data_node)
                    )
        return violations


# Helper functions


def get_provide_meshes_for_data(
        participant: ParticipantNode, data: DataNode
) -> list[MeshNode]:
    """
    This method returns all meshes provided by the given participant that use the given data.
    :param participant: The participant from which to get the meshes.
    :param data: The data that the meshes use.
    :return: A list of all meshes provided by the given participant using the given data.
    """
    provide_meshes = []
    for mesh in participant.provide_meshes:
        if data in mesh.use_data:
            provide_meshes.append(mesh)
    return provide_meshes


def filter_use_read_write_data(node) -> bool:
    """
    This method filters nodes, that could potentially use data, read data or write data.

    A mesh is the only node that can "use" data.

    A read-data node, export, watchpoint, watch-integral or action are considered to "read" data.

    A write-data- or action-node is considered to "write" data.

    Args:
         node: The node to check.

    Returns:
        True, if the node is a data-, read-/write-, action- or mesh node.
    """
    return (
            isinstance(node, DataNode)
            or isinstance(node, MeshNode)
            or isinstance(node, ReadDataNode)
            or isinstance(node, ExportNode)
            or isinstance(node, WatchPointNode)
            or isinstance(node, WatchIntegralNode)
            or isinstance(node, WriteDataNode)
            or isinstance(node, ActionNode)
    )


def filter_data_exchange(node) -> bool:
    """
    This method filters data- and exchange-nodes.
    :param node: The node to check.
    :return: True, if the node is a data- or exchange-node, False otherwise.
    """
    return isinstance(node, DataNode) or isinstance(node, ExchangeNode)


def append_participant_to_map(
        dictionary: dict[MeshNode: list[ParticipantNode]], mesh, participant
):
    """
    This method appends the given participant to the given map for an entry 'mesh'.
    If the map entry for mesh is none, then a new entry is created.
    :param dictionary: The map to append to.
    :param mesh: The mesh which entries should be appended.
    :param participant: The participant to append to the mesh.
    """
    try:
        dictionary[mesh] += [participant]
    except KeyError:
        dictionary[mesh] = [participant]


def find_parents_of_meshes(graph: Graph) -> dict[MeshNode, ParticipantNode]:
    """
    This method finds all parents (participants providing the mesh) of all meshes from the given graph.
    :param graph: The graph to find parents of meshes of.
    :return: A dictionary mapping mesh nodes to their parents.
    """
    parents_of_meshes: dict[MeshNode, ParticipantNode] = {}
    for node in graph.nodes:
        if isinstance(node, ParticipantNode):
            for mesh in node.provide_meshes:
                parents_of_meshes[mesh] = node
    return parents_of_meshes



from abc import ABC, abstractmethod

from networkx import Graph

from preciceconfigcheck.severity import Severity
from preciceconfigcheck.violation import Violation


class Rule(ABC):
    """
    Abstract Class 'Rule'. Checking a 'Rule' for violations and producing formatted output.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """@abstract property: Name of the rule in readable style."""
        pass

    def __init__(self) -> None:
        """
        Initializes a Rule object.
        """

    @abstractmethod
    def check(self, graph: Graph) -> list[Violation]:
        """
        @abstractmethod: Defines how a 'Rule' should be checked

        Args:
            graph (Graph): that must be checked according to this rule.

        Returns:
            list[Violation]: found.

        Hint: Implement Violations as inner classes in the rule of type Violation.
        """
        pass

    def satisfied(self, violations: list[Violation], debug: bool) -> bool:
        """
        Checks if a rule is satisfied.
        If debug mode is enabled, violations with severity level DEBUG are also considered.

        Args:
            violations (list[Violation]): that need to be checked.
            debug (bool): for debug mode.

        Returns:
            bool: True, if a rule is satisfied.
        """
        if debug and len(violations) > 0:
            return False
        for violation in violations:
            if violation.severity != Severity.DEBUG:
                return False
        return True



import networkx as nx
from networkx import Graph
from precice_config_graph.nodes import ParticipantNode, DataNode

from preciceconfigcheck.rule import Rule
from preciceconfigcheck.rule_utils import format_list
from preciceconfigcheck.severity import Severity
from preciceconfigcheck.violation import Violation

default_possible_solutions = [
    "Consider splitting up the simulation into multiple configurations to improve maintainability of each simulation.",
]


class DisjointSimulationsRule(Rule):
    name = "Couplings should not be disjoint"

    class CommonDisjointSimulationsViolation(Violation):
        participant_sets: frozenset[frozenset[ParticipantNode]]

        def __init__(self, participant_sets: frozenset[frozenset[ParticipantNode]]):
            super()
            assert len(participant_sets) > 1
            self.participant_sets = participant_sets

        def assemble_from_default_explanation(self, details: str = "") -> str:
            explanation = (
                f"There are {len(self.participant_sets)} simulations that do not interact with each "
                f"other{details}. "
            )

            def format_set(participants: frozenset[ParticipantNode]) -> str:
                return format_list([p.name for p in participants])

            if len(self.participant_sets) == 2:
                [participants_a, participants_b] = list(self.participant_sets)

                participants_a_label = (
                    "Participants" if len(participants_a) > 1 else "Participant"
                )
                do_str = "do" if len(participants_a) > 1 else "does"
                participants_b_label = (
                    "participants" if len(participants_b) > 1 else "participant"
                )

                explanation += (
                    f"{participants_a_label} {format_set(participants_a)} {do_str} not communicate with "
                    f"{participants_b_label} {format_set(participants_b)}."
                )
            else:
                explanation += f"Disjoint groups:"
                for component in self.participant_sets:
                    explanation += f"\n- {format_set(component)}"

            return explanation

    class FullyDisjointSimulationsViolation(CommonDisjointSimulationsViolation):
        severity = Severity.DEBUG

        def __init__(self, participant_sets: frozenset[frozenset[ParticipantNode]]):
            super().__init__(participant_sets)

        def format_explanation(self) -> str:
            return self.assemble_from_default_explanation()

        def format_possible_solutions(self) -> list[str]:
            return default_possible_solutions + [
                "Add some data to be exchanged between these simulations.",
            ]

    class SharedDataDisjointSimulationsViolation(CommonDisjointSimulationsViolation):
        severity = Severity.WARNING
        shared_data: DataNode

        def __init__(
            self,
            shared_data: DataNode,
            participant_sets: frozenset[frozenset[ParticipantNode]],
        ):
            super().__init__(participant_sets)
            self.shared_data = shared_data

        def format_explanation(self) -> str:
            return self.assemble_from_default_explanation(
                f", but share data {self.shared_data.name}"
            )

        def format_possible_solutions(self) -> list[str]:
            return default_possible_solutions + [
                "Exchange the data between these simulations.",
            ]

    def check(self, graph: Graph) -> list[Violation]:
        def is_participant(node) -> bool:
            return isinstance(node, ParticipantNode)

        def is_data_node(node) -> bool:
            return isinstance(node, DataNode)

        def has_participant(component) -> bool:
            return any(is_participant(node) for node in component)

        def get_components_with_participant(graph: Graph):
            components = list(filter(has_participant, nx.connected_components(graph)))
            # if there is just one component, remove it, since having a single component is the regular, valid case
            return components if len(components) > 1 else []

        components = list(nx.connected_components(graph))

        components_with_participant = get_components_with_participant(graph)

        # Take a look at the individual components without data, because using the same data doesn't necessarily mean
        # that the simulations really do talk to each other. In contrast to regular components, we call these dataless
        # components. It must always hold that len(components) <= len(dataless_components).
        dataless_graph = nx.subgraph_view(
            graph, filter_node=lambda node: not is_data_node(node)
        )
        dataless_components = get_components_with_participant(dataless_graph)

        fully_disjoint_participant_sets = frozenset(
            [
                frozenset(filter(is_participant, component))
                for component in components_with_participant
            ]
        )
        shared_data_participant_sets = frozenset(
            [
                frozenset(filter(is_participant, component))
                for component in dataless_components
            ]
        )

        # exclude components that get a violation for fully disjoint from shared data (fully disjoint > shared data)
        shared_data_participant_sets -= fully_disjoint_participant_sets
        fully_disjoint_participant_sets -= shared_data_participant_sets

        violations = []

        if len(fully_disjoint_participant_sets) > 1:
            violations.append(
                self.FullyDisjointSimulationsViolation(fully_disjoint_participant_sets)
            )

        dataless_components_by_data: dict[
            DataNode, set[frozenset[ParticipantNode]]
        ] = {}
        # for each dataless component, find out which data belongs to it (in order to print that information).
        for dataless_component in shared_data_participant_sets:
            # any node of a dataless component can ever only be in one overall component. So just look at the first.
            any_node = list(dataless_component)[0]
            home_component = None

            for potential_home_component in components:
                if any_node in potential_home_component:
                    home_component = potential_home_component

            assert (
                home_component is not None
            ), "Any dataless component is always in an overall component."

            # Now, use this home component to add this dataless component to one or more data names
            related_data = filter(is_data_node, home_component)
            for related_data in related_data:
                if related_data not in dataless_components_by_data:
                    dataless_components_by_data[related_data] = set()
                dataless_components_by_data[related_data].add(
                    frozenset(dataless_component)
                )

        for (data, dataless_components) in dataless_components_by_data.items():
            violations.append(
                self.SharedDataDisjointSimulationsViolation(
                    data, frozenset(dataless_components)
                )
            )

        return violations



from networkx import Graph
from precice_config_graph.nodes import (
    CouplingSchemeNode,
    MultiCouplingSchemeNode,
    ParticipantNode,
    ExchangeNode,
    MeshNode,
    MappingNode,
    DataNode,
)
import precice_config_graph.enums as e

from preciceconfigcheck.rule import Rule
from preciceconfigcheck.severity import Severity
from preciceconfigcheck.violation import Violation


class CouplingSchemeMappingRule(Rule):
    name = (
        "Exchange in a coupling scheme needs a mapping between involved participants."
    )

    class MissingMappingCouplingSchemeViolation(Violation):
        """
        A coupling scheme with a data exchange between two participants exists, but no mapping between them.
        This means they either need api-access to one-another, or they are missing a mapping.
        In case they are missing a mapping, there is this violation.
        """

        severity = Severity.ERROR

        def __init__(
            self,
            from_participant: ParticipantNode,
            to_participant: ParticipantNode,
            exchange_mesh: MeshNode,
            data: DataNode,
        ):
            self.from_participant = from_participant
            self.to_participant = to_participant
            self.exchange_mesh = exchange_mesh
            if exchange_mesh in from_participant.provide_meshes:
                # from_participant writes to own mesh, sends it to to_participant, who maps it to his own mesh, and reads from it
                self.direction = e.Direction.READ
                self.mapper = to_participant
                self.non_mapper = from_participant
                self.mesh_owner = from_participant
                self.from_string = f"from {self.exchange_mesh.name}"
                self.to_string = f"to a mesh provided by {to_participant.name}"
            else:
                # from_participant writes to own mesh, maps it to to_participants mesh, sends it to to_participant, who reads from it
                self.direction = e.Direction.WRITE
                self.mapper = from_participant
                self.non_mapper = to_participant
                self.mesh_owner = to_participant
                self.from_string = f"from a mesh provided by {from_participant.name}"
                self.to_string = f"to {self.exchange_mesh.name}"
            self.data = data

        def format_explanation(self) -> str:
            return (
                f"The exchange of data {self.data.name} belonging to the coupling scheme between participants "
                f"{self.from_participant.name} and {self.to_participant.name} using {self.mesh_owner.name}'s "
                f"mesh {self.exchange_mesh.name} is missing a mapping."
            )

        def format_possible_solutions(self) -> list[str]:
            return [
                f"For this exchange, {self.mapper.name} has to define a {self.direction.value}-mapping "
                f"{self.from_string} {self.to_string}.",
                "Otherwise, change the mesh used in the exchange and make sure that there exists a corresponding "
                "mapping for it.",
            ]

    class MissingMappingAPIAccessCouplingSchemeViolation(Violation):
        """
        A coupling scheme with a data exchange between two participants exists, but no mapping between them.
        This means they either need api-access to one-another, or they are missing a mapping.
        In case they have api-access, there is this violation, as a user can forget to specify a mapping.
        """

        severity = Severity.DEBUG

        def __init__(
            self,
            from_participant: ParticipantNode,
            to_participant: ParticipantNode,
            exchange_mesh: MeshNode,
            data: DataNode,
        ):
            self.from_participant = from_participant
            self.to_participant = to_participant
            self.exchange_mesh = exchange_mesh
            if exchange_mesh in from_participant.provide_meshes:
                # from_participant writes to own mesh, sends it to to_participant, who maps it to his own mesh, and reads from it
                self.direction = e.Direction.READ
                self.mapper = to_participant
                self.non_mapper = from_participant
                self.mesh_owner = from_participant
            else:
                # from_participant writes to own mesh, maps it to to_participants mesh, sends it to to_participant, who reads from it
                self.direction = e.Direction.WRITE
                self.mapper = from_participant
                self.non_mapper = to_participant
                self.mesh_owner = to_participant
            self.data = data

        def format_explanation(self) -> str:
            out: str = (
                f"The exchange of data {self.data.name} belonging to the coupling scheme between participants "
                f"{self.from_participant.name} and {self.to_participant.name} using {self.mesh_owner.name}'s "
                f"mesh {self.exchange_mesh.name} does not have a mapping."
            )
            out += (
                f"\nThis is valid, because {self.mapper.name} has api-access to {self.non_mapper.name}'s "
                f"{self.exchange_mesh.name}, but is more prone to errors."
            )
            return out

        def format_possible_solutions(self) -> list[str]:
            return [
                f"Specifying a just-in-time {self.direction.value}-mapping at {self.mapper.name} will make the "
                f"exchange of data more reliable."
            ]

    def check(self, graph: Graph) -> list[Violation]:
        violations: list[Violation] = []

        couplings: list[
            CouplingSchemeNode | MultiCouplingSchemeNode
        ] = filter_coupling_nodes(graph)

        for coupling in couplings:
            # Unknown participants in from=/to= are handled by precice-tools check

            # Check for all exchanges, whether a mapping exists between first and second (in the correct direction)
            exchanges: list[ExchangeNode] = coupling.exchanges
            for exchange in exchanges:
                # Check whose mesh gets used
                exchange_mesh = exchange.mesh
                from_participant = exchange.from_participant
                to_participant = exchange.to_participant
                from_meshes = from_participant.provide_meshes
                to_meshes = to_participant.provide_meshes
                data: DataNode = exchange.data

                # from-participant writes data to from-mesh, exchanges it to to-participant, who maps it to his own
                # mesh and then reads from it => READ-mapping by to-participant required, or api-access to from-mesh
                if exchange_mesh in from_meshes:
                    has_correct_mapping: bool = any(
                        mapping_fits_exchange(
                            mapping,
                            e.Direction.READ,
                            from_participant,
                            to_participant,
                            exchange_mesh,
                        )
                        for mapping in to_participant.mappings
                    )
                    # Only add a violation if no correct mapping exists
                    if not has_correct_mapping:
                        if has_api_access(to_participant, exchange_mesh):
                            # If the participant has api-access, add a debug-violation
                            violations.append(
                                self.MissingMappingAPIAccessCouplingSchemeViolation(
                                    from_participant,
                                    to_participant,
                                    exchange_mesh,
                                    data,
                                )
                            )
                        else:
                            violations.append(
                                self.MissingMappingCouplingSchemeViolation(
                                    from_participant,
                                    to_participant,
                                    exchange_mesh,
                                    data,
                                )
                            )
                # from-participant writes data to own mesh, maps it to to-mesh, exchanges it to to-participant, who
                # reads from it => WRITE-mapping by from-participant required, or api-access to to-mesh
                elif exchange_mesh in to_meshes:
                    has_correct_mapping: bool = any(
                        mapping_fits_exchange(
                            mapping,
                            e.Direction.WRITE,
                            from_participant,
                            to_participant,
                            exchange_mesh,
                        )
                        for mapping in from_participant.mappings
                    )
                    # Only add a violation if no correct mapping exists
                    if not has_correct_mapping:
                        if has_api_access(from_participant, exchange_mesh):
                            # If the participant has api-access, add a debug-violation
                            violations.append(
                                self.MissingMappingAPIAccessCouplingSchemeViolation(
                                    from_participant,
                                    to_participant,
                                    exchange_mesh,
                                    data,
                                )
                            )
                        else:
                            violations.append(
                                self.MissingMappingCouplingSchemeViolation(
                                    from_participant,
                                    to_participant,
                                    exchange_mesh,
                                    data,
                                )
                            )

        return violations


def filter_coupling_nodes(
    graph: Graph,
) -> list[CouplingSchemeNode | MultiCouplingSchemeNode]:
    """
    This function returns all coupling scheme nodes of the given graph.
    :param graph:The graph to check.
    :return: All (multi-)coupling scheme nodes of the graph.
    """
    couplings: list[CouplingSchemeNode | MultiCouplingSchemeNode] = []
    for node in graph.nodes:
        if isinstance(node, CouplingSchemeNode) or isinstance(
            node, MultiCouplingSchemeNode
        ):
            couplings.append(node)
    return couplings


def mapping_fits_exchange(
    mapping: MappingNode,
    direction: e.Direction,
    from_participant: ParticipantNode,
    to_participant: ParticipantNode,
    exchange_mesh: MeshNode,
) -> bool:
    """
        This function checks, whether the given mapping fits an exchange indicated by the given direction,
        first and second participant and mesh used in the exchange.
    :param mapping: The mapping to check.
    :param direction: The direction the mapping is supposed to have.
    :param to_participant: The to= participant in the exchange.
    :param from_participant: The from= participant in the exchange.
    :param exchange_mesh: The mesh= used in the exchange.
    :return: True, if the mapping has the correct direction and meshes.
    """
    if mapping.direction != direction:
        return False
    if direction == e.Direction.WRITE:
        # For direction-write, the mesh used in the exchange needs to be by to-participant
        if exchange_mesh not in to_participant.provide_meshes:
            return False
        if exchange_mesh != mapping.to_mesh:
            return False
    elif direction == e.Direction.READ:
        # For direction-read, the mesh used in the exchange needs to be by from-participant
        if exchange_mesh not in from_participant.provide_meshes:
            return False
        elif exchange_mesh != mapping.from_mesh:
            return False

    return True


def has_api_access(participant: ParticipantNode, mesh: MeshNode) -> bool:
    """
    This function checks, whether the given participant has API access to the given mesh.
    :param participant: The participant to check.
    :param mesh: The mesh to check.
    :return: True, if the participant receives the mesh with api-access=true
    """
    for receive_mesh in participant.receive_meshes:
        if receive_mesh.mesh == mesh and receive_mesh.api_access:
            return True
    return False



from networkx.classes import Graph
from enum import Enum
from precice_config_graph.nodes import (
    ParticipantNode,
    MeshNode,
    MappingNode,
    CouplingSchemeNode,
    MultiCouplingSchemeNode,
    ExchangeNode,
    M2NNode,
    WriteDataNode,
    ReadDataNode,
    ActionNode,
    ExportNode,
    WatchPointNode,
    WatchIntegralNode,
)
import precice_config_graph.enums as e
from preciceconfigcheck.rule_utils import rule_error_message
from preciceconfigcheck.rule import Rule
from preciceconfigcheck.severity import Severity
from preciceconfigcheck.violation import Violation


class MissingDataProcessing(Enum):
    """
    Define which data processing elements are missing.
    Data processing elements are write-data, read-data and similar elements that can write or read data.
    """

    WRITE_DATA = "write-data"
    READ_DATA = "read-data"
    READ_DATA_AND_WRITE_DATA = "read-and-write"


class MappingRule(Rule):
    name = "Mapping rules."

    class SameParticipantMappingViolation(Violation):
        """
        This class handles a mapping between two meshes of the same participant being specified.
        """

        severity = Severity.ERROR

        def __init__(
                self, participant: ParticipantNode, mesh: MeshNode, direction: e.Direction
        ):
            self.participant = participant
            self.mesh = mesh
            self.direction = direction
            if self.direction == e.Direction.READ:
                self.connecting_word = "from"
            elif self.direction == e.Direction.WRITE:
                self.connecting_word = "to"

        def format_explanation(self) -> str:
            out: str = (
                f"Participant {self.participant.name} is specifying a {self.direction.value}-mapping "
                f"{self.connecting_word} mesh {self.mesh.name}, which is his own mesh."
            )
            out += f"\nThe mapping is on {self.participant.name}'s own meshes, which is forbidden."
            return out

        def format_possible_solutions(self) -> list[str]:
            return [
                f"Please change the {self.connecting_word}-mesh to a mesh by a different participant.",
                "Otherwise, please remove the mapping to improve readability.",
            ]

    class IncorrectExchangeMappingViolation(Violation):
        """
        This class handles two participants specifying a mapping between them, but only an incorrect exchange
        alongside it, i.e., no exchange on the mesh indicated by the mapping.
        """

        severity = Severity.ERROR

        def __init__(
                self,
                parent: ParticipantNode,
                stranger: ParticipantNode,
                mesh: MeshNode,
                direction: e.Direction,
        ):
            self.parent = parent
            self.stranger = stranger
            self.mesh = mesh
            self.direction = direction
            if self.direction == e.Direction.READ:
                self.connecting_word = "from"
            elif self.direction == e.Direction.WRITE:
                self.connecting_word = "to"

        def format_explanation(self) -> str:
            out: str = (
                f"Participant {self.parent.name} is specifying a {self.direction.value}-mapping "
                f"{self.connecting_word} participant {self.stranger.name}, but the exchange in the coupling-"
                f"scheme between them is incorrect."
            )
            out += (
                f"\nFor a {self.direction.value}-mapping, {self.mesh.name} should be used to exchange "
                f"data and the participant specifying the {self.connecting_word}-mesh should be the "
                f"{self.connecting_word}-participant in the exchange."
            )
            return out

        def format_possible_solutions(self) -> list[str]:
            out: list[str] = []
            # In a 'read' mapping, the 'from' mesh is by Stranger => Stranger should be from-participant in exchange.
            if self.direction == e.Direction.READ:
                out += [
                    f"Add an exchange from {self.stranger.name} to {self.parent.name}, which uses "
                    f"{self.stranger.name}'s mesh {self.mesh.name} to the coupling scheme between them."
                ]
            # In a 'write' mapping, the 'to' mesh is by Stranger => Stranger should be to-participant in exchange.
            elif self.direction == e.Direction.WRITE:
                out += [
                    f"Add an exchange from {self.parent.name} to {self.stranger.name}, which uses "
                    f"{self.stranger.name}'s mesh {self.mesh.name}, to the coupling scheme between them."
                ]
            return out

    class MissingExchangeMappingViolation(Violation):
        """
        This class handles two participants specifying a mapping between them, but no exchange alongside it,
        i.e., no exchange between the two.
        """

        severity = Severity.ERROR

        def __init__(
                self,
                parent: ParticipantNode,
                stranger: ParticipantNode,
                mesh: MeshNode,
                direction: e.Direction,
        ):
            self.parent = parent
            self.stranger = stranger
            self.mesh = mesh
            self.direction = direction
            if self.direction == e.Direction.READ:
                self.connecting_word = "from"
            elif self.direction == e.Direction.WRITE:
                self.connecting_word = "to"

        def format_explanation(self) -> str:
            out: str = (
                f"Participant {self.parent.name} is specifying a {self.direction.value}-mapping "
                f"{self.connecting_word} participant {self.stranger.name}'s mesh {self.mesh.name}, "
                f"but there is no exchange for it in the coupling scheme between them."
            )
            out += (
                f"\nFor a {self.direction.value}-mapping, {self.mesh.name} should be used to "
                f"exchange data."
            )
            return out

        def format_possible_solutions(self) -> list[str]:
            out: list[str] = []
            # In a 'read' mapping, the 'from' mesh is by Stranger => Stranger should be from-participant in exchange.
            if self.direction == e.Direction.READ:
                out = [
                    f"Add an exchange from {self.stranger.name} to {self.parent.name}, which uses "
                    f"{self.stranger.name}'s mesh {self.mesh.name} to the coupling scheme between them."
                ]
            # In a 'write' mapping, the 'to' mesh is by Stranger => Stranger should be to-participant in exchange.
            elif self.direction == e.Direction.WRITE:
                out = [
                    f"Add an exchange from {self.parent.name} to {self.stranger.name}, which uses "
                    f"{self.stranger.name}'s mesh {self.mesh.name} to the coupling scheme between them."
                ]
            return out

    class MissingCouplingSchemeMappingViolation(Violation):
        """
        This class handles two participants specifying a mapping between them, but no coupling scheme alongside it,
        i.e., no coupling scheme exists between the two to exchange data.
        """

        severity = Severity.ERROR

        def __init__(
                self,
                parent: ParticipantNode,
                stranger: ParticipantNode,
                mesh: MeshNode,
                direction: e.Direction,
        ):
            self.parent = parent
            self.stranger = stranger
            self.mesh = mesh
            self.direction = direction
            if self.direction == e.Direction.READ:
                self.connecting_word = "from"
            elif self.direction == e.Direction.WRITE:
                self.connecting_word = "to"

        def format_explanation(self) -> str:
            return (
                f"Participant {self.parent.name} is specifying a {self.direction.value}-mapping "
                f"{self.connecting_word} participant {self.stranger.name}, but there does not exist a "
                f"coupling scheme between them."
            )

        def format_possible_solutions(self) -> list[str]:
            out: list[str] = []
            out += [
                f"Create a coupling scheme between {self.parent.name} and {self.stranger.name} with an "
                f"exchange to exchange data between them."
            ]
            out += ["Otherwise, please remove the mapping to improve readability."]
            return out

    class MissingM2NMappingViolation(Violation):
        """
        This class handles two participants specifying a mapping between them, but no m2n-exchange alongside it,
        i.e., no m2n-exchange exists between the two to exchange data.
        """

        severity = Severity.ERROR

        def __init__(
                self,
                parent: ParticipantNode,
                stranger: ParticipantNode,
                mesh: MeshNode,
                direction: e.Direction,
        ):
            self.parent = parent
            self.stranger = stranger
            self.mesh = mesh
            self.direction = direction
            if self.direction == e.Direction.READ:
                self.connecting_word = "from"
            elif self.direction == e.Direction.WRITE:
                self.connecting_word = "to"

        def format_explanation(self) -> str:
            return (
                f"The participant {self.parent.name} is specifying a {self.direction.value}-mapping "
                f"{self.connecting_word} participant {self.stranger.name}, but there does not exist an "
                f"m2n-exchange between them."
            )

        def format_possible_solutions(self) -> list[str]:
            out: list[str] = []
            out += [
                f"Create an m2n-exchange between {self.parent.name} and {self.stranger.name} to exchange data "
                f"between them."
            ]
            out += ["Otherwise, please remove the mapping to improve readability."]
            return out

    class MappingDirectionViolation(Violation):
        """
        This class handles a mapping being specified by two participants, but the to- and from-meshes not being
        according to the specified mapping direction.
        """

        severity = Severity.ERROR

        def __init__(
                self,
                parent: ParticipantNode,
                stranger: ParticipantNode,
                mesh_parent: MeshNode,
                mesh_stranger: MeshNode,
                direction: e.Direction,
        ):
            self.parent = parent
            self.stranger = stranger
            self.mesh_parent = mesh_parent
            self.mesh_stranger = mesh_stranger
            self.direction = direction
            if self.direction == e.Direction.READ:
                # X reads "from" Y
                self.connecting_word = "from"
                self.inverse_connector = "to"
                self.inverse_direction = e.Direction.WRITE
            elif self.direction == e.Direction.WRITE:
                # X writes "to" Y
                self.connecting_word = "to"
                self.inverse_connector = "from"
                self.inverse_direction = e.Direction.READ

        def format_explanation(self) -> str:
            out: str = (
                f"The {self.direction.value}-mapping of participant {self.parent.name} and mesh "
                f"{self.mesh_parent.name} {self.connecting_word} participant {self.stranger.name} and "
                f"mesh {self.mesh_stranger.name} is in the wrong direction."
            )
            out += (
                f'\nIn {self.direction.value}-mappings, the {self.inverse_connector}="mesh" has to be on a '
                f"mesh that the participant provides."
            )
            return out

        def format_possible_solutions(self) -> list[str]:
            sol: list[str] = []
            sol += [
                f'Either change direction="{self.direction.value}" to direction="{self.inverse_direction.value}", '
                f"or swap meshes {self.mesh_parent.name} and {self.mesh_stranger.name}."
            ]
            sol += [
                f"Move the mapping from {self.parent.name} to {self.stranger.name}, change its direction and "
                f"remember to update the mesh used in the <exchange .../> tag in their coupling scheme."
            ]
            sol += ["Otherwise, please remove it to improve readability."]
            return sol

    class MappingMissingDataProcessingViolation(Violation):
        """
        This class handles a "regular" mapping being specified by a participant, but no corresponding
        read- or write-data element being specified.
        """

        severity = Severity.ERROR

        def __init__(
                self,
                parent: ParticipantNode,
                stranger: ParticipantNode,
                mesh_parent: MeshNode,
                mesh_stranger: MeshNode,
                direction: e.Direction,
                missing_data_processing: MissingDataProcessing,
        ):
            self.parent = parent
            self.stranger = stranger
            self.mesh_parent = mesh_parent
            self.mesh_stranger = mesh_stranger
            self.direction = direction
            if self.direction == e.Direction.READ:
                self.connecting_word = "from"
                self.inverse_connector = "to"
                self.inverse_direction = e.Direction.WRITE
            elif self.direction == e.Direction.WRITE:
                self.connecting_word = "to"
                self.inverse_connector = "from"
                self.inverse_direction = e.Direction.READ
            self.missing_data_processing = missing_data_processing

        def format_explanation(self) -> str:
            out: str = (
                f"Participant {self.parent.name} is specifying a {self.direction.value}-mapping "
                f"{self.connecting_word} participant {self.stranger.name}'s mesh {self.mesh_stranger.name} "
                f"{self.inverse_connector} {self.parent.name}'s mesh {self.mesh_parent.name}."
            )
            if self.missing_data_processing == MissingDataProcessing.READ_DATA:
                if self.direction == e.Direction.WRITE:
                    out += (
                        f"\nHowever, it seems like {self.stranger.name} does not {self.inverse_direction.value} "
                        f"{self.inverse_connector} {self.mesh_stranger.name}."
                    )
                elif self.direction == e.Direction.READ:
                    out += (
                        f"\nHowever, it seems like {self.parent.name} does not {self.direction.value} "
                        f"{self.connecting_word} {self.mesh_parent.name}. "
                    )
            elif self.missing_data_processing == MissingDataProcessing.WRITE_DATA:
                if self.direction == e.Direction.WRITE:
                    out += (
                        f"\nHowever, it seems like {self.parent.name} does not {self.direction.value} "
                        f"{self.connecting_word} {self.mesh_parent.name}."
                    )
                elif self.direction == e.Direction.READ:
                    out += (
                        f"\nHowever, it seems like {self.stranger.name} does not {self.inverse_direction.value} "
                        f"{self.inverse_connector} {self.mesh_stranger.name}."
                    )
            elif (
                    self.missing_data_processing
                    == MissingDataProcessing.READ_DATA_AND_WRITE_DATA
            ):
                out += (
                    f"\n However, it seems like {self.parent.name} does not {self.direction.value} "
                    f"{self.connecting_word} {self.mesh_parent.name} and {self.stranger.name} does not "
                    f"{self.inverse_direction.value} {self.inverse_connector} {self.mesh_stranger.name}."
                )
            return out

        def format_possible_solutions(self) -> list[str]:
            out: list[str] = []
            if self.missing_data_processing == MissingDataProcessing.READ_DATA:
                if self.direction == e.Direction.WRITE:
                    out = [
                        f"Please ensure that {self.stranger.name} {self.inverse_direction.value}s "
                        f"{self.inverse_connector} {self.mesh_stranger.name}."
                    ]
                elif self.direction == e.Direction.READ:
                    out = [
                        f"Please ensure that {self.parent.name} {self.direction.value}s {self.connecting_word} "
                        f"{self.mesh_parent.name}."
                    ]
            elif self.missing_data_processing == MissingDataProcessing.WRITE_DATA:
                if self.direction == e.Direction.WRITE:
                    out = [
                        f"Please ensure that {self.parent.name} {self.direction.value}s {self.connecting_word} "
                        f"{self.mesh_parent.name}."
                    ]
                elif self.direction == e.Direction.READ:
                    out = [
                        f"Please ensure that {self.stranger.name} {self.inverse_direction.value}s "
                        f"{self.inverse_connector} {self.mesh_stranger.name}."
                    ]
            elif (
                    self.missing_data_processing
                    == MissingDataProcessing.READ_DATA_AND_WRITE_DATA
            ):
                out = [
                    f"Please ensure that {self.parent.name} {self.direction.value}s {self.connecting_word} "
                    f"{self.mesh_parent.name} and {self.stranger.name} {self.inverse_direction.value}s "
                    f"{self.inverse_connector} {self.mesh_stranger.name}.",
                    "Otherwise, please remove the mapping to improve readability.",
                ]
            return out

    class JustInTimeMappingApiAccessViolation(Violation):
        """
        This class handles a participant (parent) specifying a just-in-time mapping with another participant
        (stranger), but the parent not having permission to read from/write to Stranger's mesh, i.e., Parent does
        not receive the mesh with api-access=true.
        """

        severity = Severity.ERROR

        def __init__(
                self,
                parent: ParticipantNode,
                stranger: ParticipantNode,
                mesh: MeshNode,
                direction: e.Direction,
        ):
            self.parent = parent
            self.stranger = stranger
            self.mesh = mesh
            self.direction = direction
            self.connecting_word: str = ""
            if self.direction == e.Direction.READ:
                self.connecting_word = "from"
            elif self.direction == e.Direction.WRITE:
                self.connecting_word = "to"

        def format_explanation(self) -> str:
            return (
                f"The participant {self.parent.name} is specifying a just-in-time {self.direction.value}-"
                f"mapping {self.connecting_word} participant {self.stranger.name}'s mesh {self.mesh.name}, but "
                f"does not have access to it."
            )

        def format_possible_solutions(self) -> list[str]:
            return [
                f"Let {self.parent.name} receive mesh {self.mesh.name} from {self.stranger.name} with attribute "
                f'api-access="true".',
                f"Map the values from mesh {self.mesh.name} to a mesh by {self.parent.name}, before "
                f"{self.direction.value}ing them.",
                "Otherwise, please remove it to improve readability.",
            ]

    class JustInTimeMappingFormatViolation(Violation):
        """
        This class handles a just-in-time mapping being specified between two participants. For such mappings,
        only the formats read-consistent and write-conservative are allowed.
        """

        severity = Severity.ERROR

        def __init__(
                self,
                parent: ParticipantNode,
                stranger: ParticipantNode,
                mesh: MeshNode,
                direction: e.Direction,
                constraint: e.MappingConstraint,
        ):
            self.parent = parent
            self.stranger = stranger
            self.mesh = mesh
            self.direction = direction
            self.connecting_word: str = ""
            if self.direction == e.Direction.READ:
                self.connecting_word = "from"
            elif self.direction == e.Direction.WRITE:
                self.connecting_word = "to"
            self.constraint = constraint

        def format_explanation(self) -> str:
            out: str = (
                f"The just-in-time {self.direction.value}-mapping of participant {self.parent.name} "
                f"{self.connecting_word} participant {self.stranger.name}'s mesh {self.mesh.name} is in direction="
                f'"{self.direction.value}" and has constraint="{self.constraint.value}", which is invalid.'
            )
            out += (
                "\nCurrently, only just-in-time mappings of the form read-consistent and write-conservative are "
                "supported."
            )
            return out

        def format_possible_solutions(self) -> list[str]:
            out: list[str] = []
            # Mapping format is "almost" correct
            if (
                    self.direction == e.Direction.READ
                    and self.constraint == e.MappingConstraint.CONSERVATIVE
            ):
                out += [
                    f"Consider changing either the direction of the mapping between {self.parent.name} and "
                    f'{self.stranger.name} to direction="write" or the constraint of the mapping to constraint='
                    f'"consistent".'
                ]
                out += [
                    f"The effect of a read-conservative mapping can be achieved by moving the mapping from "
                    f'{self.parent.name} to {self.stranger.name} and changing its direction to "write".'
                ]
                out += [
                    f"When moving the mapping, remember to update the <exchange .../> tag in their coupling scheme."
                ]
            elif (
                    self.direction == e.Direction.WRITE
                    and self.constraint == e.MappingConstraint.CONSISTENT
            ):
                out += [
                    f"Consider changing either the direction of the mapping between {self.parent.name} and "
                    f'{self.stranger.name} to direction="read" or the constraint of the mapping to constraint='
                    f'"conservative".'
                ]
                out += [
                    f"The effect of a write-consistent mapping can be achieved by moving the mapping from "
                    f'{self.parent.name} to {self.stranger.name} and changing its direction to "read".'
                ]
                out += [
                    f"When moving the mapping, remember to update the <exchange .../> tag in their coupling scheme."
                ]
            # Generic answers for arbitrary constraints
            elif self.direction == e.Direction.READ:
                out += [
                    f"Consider changing the constraint of the mapping between {self.parent.name} and "
                    f'{self.stranger.name} from "{self.constraint.value}" to "consistent".'
                ]
            elif self.direction == e.Direction.WRITE:
                out += [
                    f"Consider changing the constraint of the mapping between {self.parent.name} and "
                    f'{self.stranger.name} from "{self.constraint.value}" to "conservative".'
                ]
            return out

    class JustInTimeMappingDirectionViolation(Violation):
        """
        This class handles a just-in-time mapping being specified between two participants, but the to- or from-mesh
        not being according to the specified mapping direction.
        """

        severity = Severity.ERROR

        def __init__(
                self,
                parent: ParticipantNode,
                stranger: ParticipantNode,
                mesh: MeshNode,
                direction: e.Direction,
        ):
            self.parent = parent
            self.stranger = stranger
            self.mesh = mesh
            self.direction = direction
            if self.direction == e.Direction.READ:
                # X reads "from" Y
                self.connecting_word = "from"
                self.inverse_direction = e.Direction.WRITE
            elif self.direction == e.Direction.WRITE:
                # X writes "to" Y
                self.connecting_word = "to"
                self.inverse_direction = e.Direction.READ

        def format_explanation(self) -> str:
            out: str = (
                f"The just-in-time {self.direction.value}-mapping of participant {self.parent.name} "
                f"{self.connecting_word} participant {self.stranger.name} and mesh {self.mesh.name} is in "
                f"the wrong direction."
            )
            out += (
                f'\nIn just-in-time{self.direction.value}-mappings, the {self.connecting_word}="mesh" has to '
                f"be on a stranger participants mesh."
            )
            return out

        def format_possible_solutions(self) -> list[str]:
            out: list[str] = []
            out += [
                f'Change direction="{self.direction.value}" to direction="{self.inverse_direction.value}".'
            ]
            out += [
                f"Move the mapping from {self.parent.name} to {self.stranger.name}, change its direction and "
                f"remember to switch the mesh used in the <exchange .../> tag in their coupling scheme."
            ]
            out += [
                f"Otherwise, change the {self.connecting_word}-mesh to a mesh by {self.stranger.name}."
            ]
            return out

    class JustInTimeMappingFormatDirectionViolation(Violation):
        """
        This class handles a just-in-time mapping being specified between two participants, but the to- or from-mesh
        not being according to the specified mapping direction and the format not being either read-consistent or
        write-conservative.
        """

        severity = Severity.ERROR

        def __init__(
                self,
                parent: ParticipantNode,
                stranger: ParticipantNode,
                mesh: MeshNode,
                direction: e.Direction,
                constraint: e.MappingConstraint,
        ):
            self.parent = parent
            self.stranger = stranger
            self.mesh = mesh
            self.direction = direction
            if self.direction == e.Direction.READ:
                self.connecting_word = "from"
                self.inverse_direction = e.Direction.WRITE
            elif self.direction == e.Direction.WRITE:
                self.connecting_word = "to"
                self.inverse_direction = e.Direction.READ
            self.constraint = constraint

        def format_explanation(self) -> str:
            out: str = (
                f'The just-in-time {self.direction.value}-mapping with constraint "{self.constraint.value}" between '
                f"participant {self.parent.name} and participant {self.stranger.name}'s mesh {self.mesh.name} has an "
                f"invalid format and is in the wrong direction."
            )
            out += (
                f'\nCurrently, only the formats "write-conservative" and "read-consistent" are implemented for '
                f"just-in-time mappings."
            )
            out += (
                f'\nIn just-in-time {self.direction.value}-mappings, the {self.connecting_word}="mesh" has to '
                f"be on a stranger participants mesh."
            )
            return out

        def format_possible_solutions(self) -> list[str]:
            out: list[str] = []
            out += [
                f'Change direction="{self.direction.value}" to direction="{self.inverse_direction.value}".'
            ]
            out += [
                f"Move the mapping from {self.parent.name} to {self.stranger.name}, change its direction and "
                f"remember to switch the mesh used in the <exchange .../> tag in their coupling scheme."
            ]
            if self.direction == e.Direction.WRITE:
                out += [
                    f"Please update the constraint of the mapping between {self.parent.name} and "
                    f'{self.stranger.name} from "{self.constraint.value}" to "conservative".'
                ]
            elif self.direction == e.Direction.READ:
                out += [
                    f"Please update the constraint of the mapping between {self.parent.name} and "
                    f'{self.stranger.name} from "{self.constraint.value}" to "consistent".'
                ]
            out += [
                f"Otherwise, change the {self.connecting_word}-mesh to a mesh by {self.stranger.name}."
            ]
            return out

    class JustInTimeMappingMethodViolation(Violation):
        """
        This class handles a just-in-time mapping being specified between two participants,
        but its mapping-method is incorrect.
        Currently, for just-time-mappings, only the mapping-methods nearest-neighbor, rbf-pum-direct and rbf are
        supported.
        """

        severity = Severity.ERROR

        def __init__(
                self,
                parent: ParticipantNode,
                stranger: ParticipantNode,
                mesh: MeshNode,
                direction: e.Direction,
                method: e.MappingMethod,
        ):
            self.parent = parent
            self.stranger = stranger
            self.mesh = mesh
            self.direction = direction
            if self.direction == e.Direction.READ:
                self.connecting_word = "from"
            elif self.direction == e.Direction.WRITE:
                self.connecting_word = "to"
            self.method = method

        def format_explanation(self) -> str:
            out: str = (
                f"The just-in-time {self.direction.value}-mapping of participant {self.parent.name} "
                f"{self.connecting_word} {self.stranger.name}'s mesh {self.mesh.name} has mapping-method "
                f'"{self.method.value}", which is invalid.'
            )
            out += (
                f'\nCurrently, only just-in-time mappings with methods "nearest-neighbor", "rbf-pum-direct" and'
                ' "rbf" are supported.'
            )
            return out

        def format_possible_solutions(self) -> list[str]:
            return [
                f'Please change the method of the mapping from "{self.method.value}" to one of the methods '
                f'"nearest-neighbor", "rbf-pum-direct" or "rbf".'
            ]

    class JustInTimeMappingMissingDataProcessingViolation(Violation):
        """
        This class handles a just-time-mapping being specified by a participant, but no corresponding
        read- or write-data element being specified.
        """

        severity = Severity.ERROR

        def __init__(
                self,
                parent: ParticipantNode,
                stranger: ParticipantNode,
                mesh: MeshNode,
                direction: e.Direction,
                missing_data_processing: MissingDataProcessing,
        ):
            self.parent = parent
            self.stranger = stranger
            self.mesh = mesh
            self.direction = direction
            if self.direction == e.Direction.READ:
                self.connecting_word = "from"
                self.inverse_connector = "to"
                self.inverse_direction = e.Direction.WRITE
            elif self.direction == e.Direction.WRITE:
                self.connecting_word = "to"
                self.inverse_connector = "from"
                self.inverse_direction = e.Direction.READ
            self.missing_data_processing = missing_data_processing

        def format_explanation(self) -> str:
            out: str = (
                f"Participant {self.parent.name} is specifying a just-in-time {self.direction.value}-mapping "
                f"{self.connecting_word} participant {self.stranger.name}'s mesh {self.mesh.name}. "
            )
            if self.missing_data_processing == MissingDataProcessing.READ_DATA:
                if self.direction == e.Direction.WRITE:
                    out += (
                        f"\nHowever, it seems like {self.stranger.name} does not {self.inverse_direction.value} "
                        f"{self.inverse_connector} {self.mesh.name}."
                    )
                elif self.direction == e.Direction.READ:
                    out += (
                        f"\nHowever, it seems like {self.parent.name} does not {self.direction.value} "
                        f"{self.connecting_word} {self.mesh.name}."
                    )
            elif self.missing_data_processing == MissingDataProcessing.WRITE_DATA:
                if self.direction == e.Direction.WRITE:
                    out += (
                        f"\nHowever, it seems like {self.parent.name} does not {self.inverse_direction.value} "
                        f"{self.inverse_connector} {self.mesh.name}."
                    )
                elif self.direction == e.Direction.READ:
                    out += (
                        f"\nHowever, it seems like {self.stranger.name} does not {self.direction.value} "
                        f"{self.connecting_word} {self.mesh.name}."
                    )
            elif (
                    self.missing_data_processing
                    == MissingDataProcessing.READ_DATA_AND_WRITE_DATA
            ):
                out += (
                    f"\nHowever, it seems like {self.parent.name} does not {self.direction.value} "
                    f"{self.connecting_word} {self.mesh.name} and {self.stranger.name} does not "
                    f"{self.inverse_direction.value} {self.inverse_connector} {self.mesh.name}."
                )
            return out

        def format_possible_solutions(self) -> list[str]:
            out: list[str] = []
            if self.missing_data_processing == MissingDataProcessing.READ_DATA:
                if self.direction == e.Direction.WRITE:
                    out = [
                        f"Please ensure that {self.stranger.name} {self.inverse_direction.value}s "
                        f"{self.inverse_connector} {self.mesh.name}."
                    ]
                elif self.direction == e.Direction.READ:
                    out = [
                        f"Please ensure that {self.parent.name} {self.direction.value}s "
                        f"{self.connecting_word} {self.mesh.name}."
                    ]
            elif self.missing_data_processing == MissingDataProcessing.WRITE_DATA:
                if self.direction == e.Direction.WRITE:
                    out = [
                        f"Please ensure that {self.parent.name} {self.inverse_direction.value}s "
                        f"{self.inverse_connector} {self.mesh.name}."
                    ]
                elif self.direction == e.Direction.READ:
                    out = [
                        f"Please ensure that {self.stranger.name} {self.direction.value}s "
                        f"{self.connecting_word} {self.mesh.name}."
                    ]
            elif (
                    self.missing_data_processing
                    == MissingDataProcessing.READ_DATA_AND_WRITE_DATA
            ):
                out = [
                    f"Please ensure that {self.parent.name} {self.direction.value}s {self.connecting_word} "
                    f"{self.mesh.name} and {self.stranger.name} {self.inverse_direction.value}s "
                    f"{self.inverse_connector} {self.mesh.name}.",
                    "Otherwise, please remove the mapping to improve readability.",
                ]
            return out

    def check(self, graph: Graph) -> list[Violation]:
        violations: list[Violation] = []

        m2ns: list[M2NNode] = filter_m2n_nodes(graph)
        coupling_schemes: list[
            CouplingSchemeNode | MultiCouplingSchemeNode
            ] = filter_coupling_nodes(graph)

        mappings: list[MappingNode] = filter_mapping_nodes(graph)
        for mapping in mappings:
            method: e.MappingMethod = mapping.method
            direction: e.Direction = mapping.direction
            constraint: e.MappingConstraint = mapping.constraint
            participant_parent: ParticipantNode = mapping.parent_participant
            # Initialize to avoid keeping value for multiple iterations (mappings)
            mesh_parent: MeshNode = None
            mesh_stranger: MeshNode = None

            # Determine which tags exist and thus which meshes belong to which participant
            if mapping.just_in_time:
                # Either 'to' or 'from' tag does not exist in JIT mappings
                if mapping.from_mesh and not mapping.to_mesh:
                    mesh_stranger = mapping.from_mesh
                elif not mapping.from_mesh and mapping.to_mesh:
                    mesh_stranger = mapping.to_mesh
                else:
                    rule_error_message('Mapping must have attribute "to" or "from"')

            # "regular" mappings
            else:
                if mapping.from_mesh in participant_parent.provide_meshes:
                    mesh_parent = mapping.from_mesh
                    mesh_stranger = mapping.to_mesh
                elif mapping.to_mesh in participant_parent.provide_meshes:
                    mesh_parent = mapping.to_mesh
                    mesh_stranger = mapping.from_mesh
                else:
                    rule_error_message(
                        "One mesh in mapping must be provided by parent participant"
                    )

            # These violations are the same for both JIT and regular mappings
            participant_strangers = get_participants_of_mesh(graph, mesh_stranger)

            if len(participant_strangers) == 0:
                # This mapping is broken
                # mesh_stranger is not provided by any participant
                # Gets handled in provide_mesh.py
                continue

            if len(participant_strangers) > 1:
                # This mapping is broken
                # mesh_stranger gets claimed by more than one participant
                # Gets handled in provide_mesh.py
                continue

            # There is only one participant claiming the mesh
            participant_stranger = participant_strangers[0]

            if participant_parent == participant_stranger:
                # This mapping is broken
                # The mapping is between meshes of the same participant
                violations.append(
                    self.SameParticipantMappingViolation(
                        participant_parent, mesh_stranger, direction
                    )
                )
                continue

            # Check JIT mapping specific violations
            if mapping.just_in_time:
                # Only the methods nearest-neighbor, rbf-pum-direct and rbf are supported
                supported_methods = [
                    e.MappingMethod.NEAREST_NEIGHBOR,
                    e.MappingMethod.RBF_PUM_DIRECT,
                    e.MappingMethod.RBF,
                ]
                if method not in supported_methods:
                    violations.append(
                        self.JustInTimeMappingMethodViolation(
                            participant_parent,
                            participant_stranger,
                            mesh_stranger,
                            direction,
                            method,
                        )
                    )

                # Only read-consistent and write-conservative are supported
                if direction == e.Direction.READ:
                    # For a read-mapping, the 'from' mesh needs to be defined
                    if mapping.from_mesh:
                        # Correct direction
                        # For a JIT-read-mapping, the constraint has to be consistent
                        if constraint != e.MappingConstraint.CONSISTENT:
                            # Correct direction, but wrong constraint
                            violations.append(
                                self.JustInTimeMappingFormatViolation(
                                    participant_parent,
                                    participant_stranger,
                                    mesh_stranger,
                                    direction,
                                    constraint,
                                )
                            )
                    else:
                        # Wrong direction
                        if constraint == e.MappingConstraint.CONSISTENT:
                            # Correct format
                            violations.append(
                                self.JustInTimeMappinge.DirectionViolation(
                                    participant_parent,
                                    participant_stranger,
                                    mesh_stranger,
                                    direction,
                                )
                            )
                        else:
                            # Wrong format and wrong direction
                            violations.append(
                                self.JustInTimeMappingFormatDirectionViolation(
                                    participant_parent,
                                    participant_stranger,
                                    mesh_stranger,
                                    direction,
                                    constraint,
                                )
                            )
                elif direction == e.Direction.WRITE:
                    # For a write-mapping, the 'to' mesh needs to be defined
                    if mapping.to_mesh:
                        # Correct direction
                        # For a JIT-write-mapping, the constraint has to be conservative
                        if constraint != e.MappingConstraint.CONSERVATIVE:
                            # Correct direction, but wrong constraint
                            violations.append(
                                self.JustInTimeMappingFormatViolation(
                                    participant_parent,
                                    participant_stranger,
                                    mesh_stranger,
                                    direction,
                                    constraint,
                                )
                            )
                    else:
                        # Wrong direction
                        if constraint == e.MappingConstraint.CONSERVATIVE:
                            # Correct format
                            violations.append(
                                self.JustInTimeMappinge.DirectionViolation(
                                    participant_parent,
                                    participant_stranger,
                                    mesh_stranger,
                                    direction,
                                )
                            )
                        else:
                            # Wrong format and wrong direction
                            violations.append(
                                self.JustInTimeMappingFormate.DirectionViolation(
                                    participant_parent,
                                    participant_stranger,
                                    mesh_stranger,
                                    direction,
                                    constraint,
                                )
                            )

                # Check if participant receives mesh with api-access="true"
                receive_meshes = mapping.parent_participant.receive_meshes
                for receive_mesh in receive_meshes:
                    if receive_mesh.mesh == mesh_stranger:
                        # If api-access != true, then participant does not have permission to read from/write to it
                        if not receive_mesh.api_access:
                            violations.append(
                                self.JustInTimeMappingApiAccessViolation(
                                    participant_parent,
                                    participant_stranger,
                                    mesh_stranger,
                                    direction,
                                )
                            )
                write_datas = []
                read_datas = []
                # Check if there are corresponding read-/write-data elements for the mapping
                if direction == e.Direction.WRITE:
                    write_datas = get_write_datas_of_mesh(
                        mesh_stranger, participant_parent
                    )
                    read_datas = get_read_datas_of_mesh(
                        mesh_stranger, participant_stranger
                    )
                elif direction == e.Direction.READ:
                    read_datas = get_read_datas_of_mesh(
                        mesh_stranger, participant_parent
                    )
                    write_datas = get_write_datas_of_mesh(
                        mesh_stranger, participant_stranger
                    )
                    # In a JIT write-mapping, Parent should write to Stranger's mesh; Stranger should read from their own mesh
                    # In a JIT read-mapping, Parent should read from Stranger's mesh; Stranger should write to their own mesh
                if len(write_datas) == 0 and len(read_datas) == 0:
                    violations.append(
                        self.JustInTimeMappingMissingDataProcessingViolation(
                            participant_parent,
                            participant_stranger,
                            mesh_stranger,
                            direction,
                            MissingDataProcessing.READ_DATA_AND_WRITE_DATA,
                        )
                    )
                elif len(write_datas) != 0 and len(read_datas) == 0:
                    violations.append(
                        self.JustInTimeMappingMissingDataProcessingViolation(
                            participant_parent,
                            participant_stranger,
                            mesh_stranger,
                            direction,
                            MissingDataProcessing.READ_DATA,
                        )
                    )
                elif len(write_datas) == 0 and len(read_datas) != 0:
                    violations.append(
                        self.JustInTimeMappingMissingDataProcessingViolation(
                            participant_parent,
                            participant_stranger,
                            mesh_stranger,
                            direction,
                            MissingDataProcessing.WRITE_DATA,
                        )
                    )
            # Only regular mappings here
            else:
                if direction == e.Direction.READ:
                    # In a read-mapping, the 'to'-mesh has to be by the parent, the 'from'-mesh by Stranger
                    if not mesh_parent == mapping.to_mesh:
                        violations.append(
                            self.Mappinge.DirectionViolation(
                                participant_parent,
                                participant_stranger,
                                mesh_parent,
                                mesh_stranger,
                                direction,
                            )
                        )
                elif direction == e.Direction.WRITE:
                    # In a write-mapping, the 'from'-mesh has to be by the parent, the 'to'-mesh by stranger
                    if not mesh_parent == mapping.from_mesh:
                        violations.append(
                            self.MappingDirectionViolation(
                                participant_parent,
                                participant_stranger,
                                mesh_parent,
                                mesh_stranger,
                                direction,
                            )
                        )
                write_datas = []
                read_datas = []
                # Check if there are corresponding read-/write-data elements for the mapping
                if direction == e.Direction.WRITE:
                    # In a write-mapping, Parent should write to their own mesh; Stranger should read from their own mesh
                    write_datas = get_write_datas_of_mesh(
                        mesh_parent, participant_parent
                    )
                    read_datas = get_read_datas_of_mesh(
                        mesh_stranger, participant_stranger
                    )
                elif direction == e.Direction.READ:
                    # In a read-mapping, Parent should read from their own mesh; Stranger should write to their own
                    write_datas = get_write_datas_of_mesh(
                        mesh_stranger, participant_stranger
                    )
                    read_datas = get_read_datas_of_mesh(mesh_parent, participant_parent)
                if len(read_datas) == 0 and len(write_datas) == 0:
                    # only read is missing
                    violations.append(
                        self.MappingMissingDataProcessingViolation(
                            participant_parent,
                            participant_stranger,
                            mesh_parent,
                            mesh_stranger,
                            direction,
                            MissingDataProcessing.READ_DATA_AND_WRITE_DATA,
                        )
                    )
                elif len(read_datas) == 0 and len(write_datas) != 0:
                    violations.append(
                        self.MappingMissingDataProcessingViolation(
                            participant_parent,
                            participant_stranger,
                            mesh_parent,
                            mesh_stranger,
                            direction,
                            MissingDataProcessing.READ_DATA,
                        )
                    )
                elif len(read_datas) != 0 and len(write_datas) == 0:
                    violations.append(
                        self.MappingMissingDataProcessingViolation(
                            participant_parent,
                            participant_stranger,
                            mesh_parent,
                            mesh_stranger,
                            direction,
                            MissingDataProcessing.WRITE_DATA,
                        )
                    )
            # Both JIT and regular mappings share these violations
            # Check for correct m2n between mapping participants
            m2n = get_m2n_of_participants(
                m2ns, participant_parent, participant_stranger
            )
            if not m2n:
                violations.append(
                    self.MissingM2NMappingViolation(
                        participant_parent,
                        participant_stranger,
                        mesh_stranger,
                        direction,
                    )
                )

            # Check if mapping-participants also have a coupling scheme and then also an exchange
            couplings = get_coupling_schemes_of_mapping(
                coupling_schemes, participant_parent, participant_stranger
            )
            if len(couplings) == 0:
                # Participants try to map data, but there exists no coupling scheme between them
                violations.append(
                    self.MissingCouplingSchemeMappingViolation(
                        participant_parent,
                        participant_stranger,
                        mesh_stranger,
                        direction,
                    )
                )
                # Continue with the next mapping; this one cannot cause more violations
                continue

            # Both JIT and regular mappings
            exchanges = get_exchange_of_participants(
                participant_parent, participant_stranger, couplings
            )
            if len(exchanges) == 0:
                # Participants try to map data, and there exists a coupling-scheme between them, but not any exchanges
                violations.append(
                    self.MissingExchangeMappingViolation(
                        participant_parent,
                        participant_stranger,
                        mesh_stranger,
                        direction,
                    )
                )
                # Continue with the next mapping; this one cannot cause more violations
                continue
            # Check if any exchange of the participants coupling scheme is “correct”, i.e., on the strangers mesh
            if not any(
                    exchange_belongs_to_mapping(
                        exchange, direction, participant_stranger, mesh_stranger
                    )
                    for exchange in exchanges
            ):
                # No correct exchange, i.e., using correct mesh
                violations.append(
                    self.IncorrectExchangeMappingViolation(
                        participant_parent,
                        participant_stranger,
                        mesh_stranger,
                        direction,
                    )
                )

        return violations


# Helper functions
def get_m2n_of_participants(
        m2ns: list[M2NNode], participant_a: ParticipantNode, participant_b: ParticipantNode
) -> M2NNode | None:
    """
    This function returns the m2n-node connecting two participants if one exists.
    :param m2ns: All m2n nodes that get searched.
    :param participant_a: The first participant for which the m2n node is needed.
    :param participant_b: The other participant.
    :return: The m2n node between the given participants if it exists, None otherwise.
    """
    for m2n in m2ns:
        if m2n.acceptor == participant_a and m2n.connector == participant_b:
            return m2n
        elif m2n.acceptor == participant_b and m2n.connector == participant_a:
            return m2n
    return None


def exchange_belongs_to_mapping(
        exchange: ExchangeNode,
        direction: e.Direction,
        participant_stranger: ParticipantNode,
        mesh_stranger: MeshNode,
) -> bool:
    """
    Evaluates whether the given exchange is correct for mapping corresponding to the given (stranger) participant,
    his mesh and the direction.

    It is known that the exchange belongs to both participants of the mapping and the given participant is NOT the
    parent of the mapping. Then, participant_stranger, mesh_stranger and direction should be enough to identify
    whether the exchange belongs to the mapping or not.
    :param exchange: The exchange to check.
    :param direction: The direction of the mapping to check.
    :param participant_stranger: The participant which is needed to check whether the exchange belongs to the mapping
    :return: True if the exchange is in the correct direction and uses the correct mesh, False otherwise.
    """
    ex_dir: bool = False
    ex_mesh: bool = False
    if direction == e.Direction.READ:
        # In a read-mapping, Stranger should correspond to the exchange's from-participant
        if participant_stranger == exchange.from_participant:
            ex_dir = True
            # Otherwise, the direction is false
    elif direction == e.Direction.WRITE:
        # In a write-mapping, Stranger should correspond to the exchange's to-participant
        if participant_stranger == exchange.to_participant:
            ex_dir = True
            # Otherwise, the direction is false
    # In a mapping, Stranger's mesh should be used
    if mesh_stranger == exchange.mesh:
        ex_mesh = True
        # Otherwise, the mesh used is incorrect
    return ex_dir and ex_mesh


def get_participants_of_mesh(graph: Graph, mesh: MeshNode) -> list[ParticipantNode]:
    """
    This method returns the participant who owns the given mesh.
    :param graph: The graph of the preCICE config.
    :param mesh: The mesh of which the participant is needed.
    :return: The participant who owns the mesh.
    """
    participants: list[ParticipantNode] = []
    for node in graph.nodes:
        if isinstance(node, ParticipantNode):
            if mesh in node.provide_meshes:
                participants.append(node)
    # If participant does not exist, this will lead to a violation later
    return participants


def get_coupling_schemes_of_mapping(
        coupling_schemes: list[CouplingSchemeNode | MultiCouplingSchemeNode],
        participant_a: ParticipantNode,
        participant_b: ParticipantNode,
) -> list[CouplingSchemeNode | MultiCouplingSchemeNode]:
    """
    This method returns the coupling scheme between the given participants.
    :param coupling_schemes: All coupling schemes of the preCICE config.
    :param participant_a: One of the participants for which the coupling scheme is needed.
    :param participant_b: The other participant.
    :return:The coupling scheme between the participants; None if there is no coupling scheme between the participants.
    """
    couplings: list[CouplingSchemeNode | MultiCouplingSchemeNode] = []
    for coupling in coupling_schemes:
        if isinstance(coupling, CouplingSchemeNode):
            if (
                    participant_a == coupling.first_participant
                    and participant_b == coupling.second_participant
            ) or (
                    participant_b == coupling.first_participant
                    and participant_a == coupling.second_participant
            ):
                couplings += [coupling]
        elif isinstance(coupling, MultiCouplingSchemeNode):
            if (
                    participant_a in coupling.participants
                    and participant_b in coupling.participants
            ):
                couplings += [coupling]
    return couplings


def get_exchange_of_participants(
        participant_a: ParticipantNode,
        participant_b: ParticipantNode,
        couplings: list[CouplingSchemeNode | MultiCouplingSchemeNode],
) -> list[ExchangeNode]:
    """
    This method returns the exchange nodes between the given participants in the given coupling scheme if any exist.
    :param participant_a: One of the participants for which the exchange is needed.
    :param participant_b: The other participant.
    :param couplings: The coupling scheme(s) between the participants, in which the exchange should exist.
    :return: A list of exchange nodes between the two participants, if any exist, else None.
    """
    exchange_nodes: list[ExchangeNode] = []
    for coupling in couplings:
        # Both coupling scheme nodes and multi-coupling-scheme-nodes have the same attribute 'exchanges'
        for exchange in coupling.exchanges:
            if (
                    participant_a == exchange.to_participant
                    and participant_b == exchange.from_participant
            ) or (
                    participant_b == exchange.to_participant
                    and participant_a == exchange.from_participant
            ):
                exchange_nodes.append(exchange)
    return exchange_nodes


def filter_mapping_nodes(graph: Graph) -> list[MappingNode]:
    """
    This function returns all mapping nodes of the given graph.
    :param graph:The graph to check.
    :return: All mapping nodes of the graph.
    """
    mappings: list[MappingNode] = []
    for node in graph.nodes:
        if isinstance(node, MappingNode):
            mappings.append(node)
    return mappings


def filter_coupling_nodes(
        graph: Graph,
) -> list[CouplingSchemeNode | MultiCouplingSchemeNode]:
    """
    This function returns all coupling scheme nodes of the given graph.
    :param graph:The graph to check.
    :return: All (multi-)coupling scheme nodes of the graph.
    """
    couplings: list[CouplingSchemeNode | MultiCouplingSchemeNode] = []
    for node in graph.nodes:
        if isinstance(node, CouplingSchemeNode) or isinstance(
                node, MultiCouplingSchemeNode
        ):
            couplings.append(node)
    return couplings


def filter_parallel_coupling_nodes(
        graph: Graph,
) -> list[CouplingSchemeNode | MultiCouplingSchemeNode]:
    """
    This function returns all parallel coupling scheme nodes of the given graph.
    :param graph:The graph to check.
    :return: All coupling scheme nodes of the graph, which are either of type multi or ...-parallel.
    """
    couplings: list[CouplingSchemeNode | MultiCouplingSchemeNode] = []
    for node in graph.nodes:
        if isinstance(node, CouplingSchemeNode):
            if node.type in [
                e.CouplingSchemeType.PARALLEL_EXPLICIT,
                e.CouplingSchemeType.PARALLEL_IMPLICIT,
            ]:
                couplings.append(node)
        elif isinstance(node, MultiCouplingSchemeNode):
            couplings.append(node)
    return couplings


def filter_m2n_nodes(graph: Graph) -> list[M2NNode]:
    """
    This function returns all m2n nodes of the given graph.
    :param graph:The graph to check.
    :return: All m2n nodes of the graph.
    """
    m2ns: list[M2NNode] = []
    for node in graph.nodes:
        if isinstance(node, M2NNode):
            m2ns.append(node)
    return m2ns


def get_write_datas_of_mesh(
        mesh: MeshNode, parent: ParticipantNode
) -> list[WriteDataNode | ActionNode]:
    """
    Get all writers of the given mesh, which gets provided by parent.
    Data can be written by a write-data node or an action node.
    :param mesh: The mesh to get writers for.
    :param parent: The participant providing mesh.
    :return: A list of all writers of the given mesh, if any.
    """
    writers: list[WriteDataNode | ActionNode] = []
    for action in parent.actions:
        if action.mesh == mesh:
            writers.append(action)
    for write_data in parent.write_data:
        if write_data.mesh == mesh:
            writers.append(write_data)
    return writers


def get_read_datas_of_mesh(
        mesh: MeshNode, parent: ParticipantNode
) -> list[ReadDataNode | ExportNode | ActionNode | WatchPointNode | WatchIntegralNode]:
    """
    Get all readers of the given mesh, which gets provided by parent.
    Data can be read by a read-data node, an export node, an action node, or a watch-point /-integral node.
    :param mesh: The mesh to get writers for.
    :param parent: The participant providing mesh.
    :return: A list of all readers of the given mesh, if any.
    """
    readers: list[
        ReadDataNode | ExportNode | ActionNode | WatchPointNode | WatchIntegralNode
        ] = []
    # Export “reads” all meshes
    for export in parent.exports:
        readers.append(export)
    for action in parent.actions:
        if action.mesh == mesh:
            readers.append(action)
    for read_data in parent.read_data:
        if read_data.mesh == mesh:
            readers.append(read_data)
    for watchpoint in parent.watchpoints:
        if watchpoint.mesh == mesh:
            readers.append(watchpoint)
    for watch_integral in parent.watch_integrals:
        if watch_integral.mesh == mesh:
            readers.append(watch_integral)
    return readers



from abc import ABC, abstractmethod

from preciceconfigcheck.severity import Severity
import preciceconfigcheck.color as c


class Violation(ABC):
    """
    Abstract Class 'Violation'. Creates a formatted string for its attributes.
    """

    line: int = None
    """Attribute: Do not use 'Violation.line', use 'self.line'"""

    @property
    @abstractmethod
    def severity(self) -> Severity:
        """@abstract property: Type"""
        pass

    @abstractmethod
    def __init__(self, line: int) -> None:
        """
        @abstractmethod: Initializes a 'Violation' object.

        Args:
            line (int): The line in the config.xml file of the violation.

        Hint: When overwriting, it is recommended to pass on appropriate attributes.
        Later, these attributes can be called with 'self.attribute'.
        """
        self.line = line

    @abstractmethod
    def format_explanation(self) -> str:
        """
        @abstractmethod: Formats the explanation of 'Violation'.

        Returns:
            str: formatted explanation

        Hint: Use the attributes defined in '__init__()'.
        """
        pass

    @abstractmethod
    def format_possible_solutions(self) -> list[str]:
        """
        @abstractmethod: Formats multiple possible solutions of 'Violation'.

        Returns:
            list[str]: of formatted possible solutions

        Hint: Use the attributes defined in '__init__()'.
        """
        pass

    def format(self, debug: bool) -> str | None:
        """
        Formats the 'Violation' for its attributes.
        If debug mode is enabled, violations with DEBUG severity are also formatted.

        Args:
            debug (bool): for debug mode.

        Returns:
            str: formatted 'Violation'
        """
        if not debug and (self.severity == Severity.DEBUG):
            return None

        severity_info: str = f"[{self.severity.value}]: "
        class_name: str = (
            f"({c.dyeing(self.__class__.__name__, c.purple)}) " if debug else ""
        )
        existing_line: str = f"(Line {self.line}) " if self.line else ""
        # indent additional lines of the explanation to be aligned with first row after ">>> " is added
        explanation: str = self.format_explanation().replace("\n", "\n     ")
        possible_solutions: list[str] = self.format_possible_solutions()

        out: str = (
            c.dyeing(" >>> ", c.cyan)
            + severity_info
            + class_name
            + existing_line
            + explanation
        )
        for possible_solution in possible_solutions:
            out += c.dyeing("\n     ==> ", c.cyan) + possible_solution
        return out



from networkx import Graph

import preciceconfigcheck.color as c
from preciceconfigcheck.rule import Rule
from preciceconfigcheck.rules import (
    missing_coupling,
    missing_exchange,
    data_use_read_write,
    compositional_coupling,
    mapping,
    m2n_exchange,
    disjoint_simulations,
    provide_mesh,
    coupling_scheme_mapping,
    receive_mesh,
)
from preciceconfigcheck.severity import Severity
from preciceconfigcheck.violation import Violation

rules: list[Rule] = [
    missing_coupling.MissingCouplingSchemeRule(),
    missing_exchange.MissingExchangeRule(),
    data_use_read_write.DataUseReadWriteRule(),
    compositional_coupling.CompositionalCouplingRule(),
    m2n_exchange.M2NExchangeRule(),
    mapping.MappingRule(),
    disjoint_simulations.DisjointSimulationsRule(),
    provide_mesh.ProvideMeshRule(),
    coupling_scheme_mapping.CouplingSchemeMappingRule(),
    receive_mesh.ReceiveMeshRule(),
]


def has_satisfied_rules(
    violations_by_rule: dict[Rule, list[Violation]], debug: bool
) -> bool:
    """
    Checks if at least one rule is satisfied.
    If debug mode is enabled, violations with severity level DEBUG are also considered.

    Args:
        violations_by_rule (dict[Rule, list[Violation]]): of rules and their violations that need to be checked.
        debug (bool): for debug mode.

    Returns:
        bool: True, if at least one rule is satisfied.
    """
    for (rule, violations) in violations_by_rule.items():
        if rule.satisfied(violations, debug):
            return True
    return False


def has_unsatisfied_rules(violations_by_rule: dict[Rule, list[Violation]], debug: bool):
    """
    Checks if at least one rule is unsatisfied.
    If debug mode is enabled, violations with severity level DEBUG are also considered.

    Args:
        violations_by_rule (dict[Rule, list[Violation]]): of rules and their violations that need to be checked.
        debug (bool): for debug mode.

    Returns:
        bool: True, if at least one rule is unsatisfied.
    """
    for (rule, violations) in violations_by_rule.items():
        if not rule.satisfied(violations, debug):
            return True
    return False


def check_all_rules(graph: Graph, debug: bool) -> dict[Rule, list[Violation]]:
    """
    Checks a graph for violations according to the rules.
    If debug mode is enabled, violations with severity level DEBUG are also considered.

    Args:
        graph (Graph): that needs to be checked with all rules.
        debug (bool): for debug mode.

    Returns:
        dict[Rule, list[Violation]]: of rules and their violations.
    """
    print("")
    if debug:
        print("Checking rules...")

    violations_by_rule = {}
    for rule in rules:
        violations_by_rule[rule] = rule.check(graph)

    if debug:
        print("Rules checked.\n")

    return violations_by_rule


def print_all_results(
    violations_by_rule: dict[Rule, list[Violation]], debug: bool
) -> None:
    """
    Prints all existing violations of all rules.
    If debug mode is enabled, violations with severity level DEBUG are also considered.

    Args:
        violations_by_rule (dict[Rule, list[Violation]]): of rules and their violations that must be printed.
        debug (bool): for debug mode.

    Returns:
        None
    """

    def total_violations_by_severity(severity: Severity) -> int:
        """
        Counts how many violations of a severity type there are.

        Args:
            severity (Severity): type that must be counted.

        Returns:
            int: Number of severity type.
        """
        result: int = 0
        for violations in violations_by_rule.values():
            for violation in violations:
                if violation.severity == severity:
                    result += 1
        return result

    if has_satisfied_rules(violations_by_rule, debug) and debug:
        print(f"[{Severity.DEBUG.value}]: These rules are satisfied:")

        for (rule, violations) in violations_by_rule.items():
            if rule.satisfied(violations, debug):
                print_result(rule, violations, debug)

        print("")

    if has_unsatisfied_rules(violations_by_rule, debug):
        print("The following issues were found:\n")

        for (rule, violations) in violations_by_rule.items():
            if not rule.satisfied(violations, debug):
                print_result(rule, violations, debug)

        total_num_errors = total_violations_by_severity(Severity.ERROR)
        total_num_warnings = total_violations_by_severity(Severity.WARNING)
        error_str: str = Severity.ERROR.value + ("s" if total_num_errors != 1 else "")
        warning_str: str = Severity.WARNING.value + (
            "s" if total_num_warnings != 1 else ""
        )
        print(
            f"Your configuration file raised {total_num_errors} {error_str} "
            f"and {total_num_warnings} {warning_str}."
        )
        print("Please review your configuration file before continuing.")


def print_result(rule: Rule, violations: list[Violation], debug: bool) -> None:
    """
    If the rule has violations, these will be printed.
    If debug mode is enabled, more information is displayed.

    Args:
        rule (Rule): that must be printed.
        violations (list[Violation]): that must be printed.
        debug (bool): for debug mode.

    Returns:
        None
    """
    if rule.satisfied(violations, debug):
        if debug:
            print(f" - {c.dyeing(rule.__class__.__name__, c.purple)}")
        return
    else:
        rule_name: str = (
            f"({c.dyeing(rule.__class__.__name__, c.purple)}) {rule.name}"
            if debug
            else f"{rule.name}"
        )
        print(rule_name)

        for violation in violations:
            formatted_violation = violation.format(debug)
            if formatted_violation:
                print(formatted_violation)

        print("")
